import itertools
from numbers import Number
import tempfile
from typing import Any, Callable, Generic, TypeVar, Generator

import pydot
import sympy

from joblib import Parallel, delayed
import joblib
import sys
import os
from tqdm import tqdm
import numpy as np

__all__ = [
    "set_n_parallel_jobs",
    "get_n_parallel_jobs",
    "is_using_parallel_processing",
    "parallel",
    "delayed",
]

PARALLELIZE = True
N_PARALLEL_PROCESSES = os.cpu_count()

NUMPY_FLOAT_TYPE = np.float32


def _lambdify_type_check(*args, **kwargs):
    f = sympy.lambdify(*args, **kwargs)

    def f_type_checked(*args, **kwargs):
        for a in args:
            if isinstance(a, np.ndarray):
                if a.dtype != NUMPY_FLOAT_TYPE:
                    raise ValueError(f"Expected {NUMPY_FLOAT_TYPE}, got {a.dtype}")
            elif not isinstance(a, Number):
                raise ValueError(f"Expected {NUMPY_FLOAT_TYPE}, got {type(a)}")
        for v in kwargs.values():
            if isinstance(v, np.ndarray):
                if v.dtype != NUMPY_FLOAT_TYPE:
                    raise ValueError(f"Expected {NUMPY_FLOAT_TYPE}, got {v.dtype}")
            elif not isinstance(v, Number):
                raise ValueError(f"Expected {NUMPY_FLOAT_TYPE}, got {type(v)}")
        return f(*args, **kwargs)

    return f_type_checked


def set_n_parallel_jobs(n_jobs: int, print_message: bool = False) -> None:
    """
    Set the number of parallel jobs to use.

    Parameters
    ----------
    n_jobs : int
        The number of parallel jobs to use.
    print_message : bool, optional
        Whether to print a message when the number of parallel jobs is set.
    """
    global N_PARALLEL_PROCESSES
    N_PARALLEL_PROCESSES = n_jobs
    global PARALLELIZE
    PARALLELIZE = n_jobs > 1
    if print_message:
        print(f"Using {n_jobs} parallel job{'s' if n_jobs > 1 else ''}")


def get_n_parallel_jobs() -> int:
    """
    Returns the number of parallel jobs being used. If parallel processing is not
    enabled, returns 1.
    """
    return N_PARALLEL_PROCESSES if is_using_parallel_processing() else 1


def is_using_parallel_processing() -> bool:
    """Returns True if parallel processing is enabled."""
    return PARALLELIZE and N_PARALLEL_PROCESSES > 1


def _expfmt(x):
    if isinstance(x, Number):
        x = round(x)
        if x < 10000:
            return f"{x}"
        x = f"{x:.2e}"
    else:
        x = str(x)
    if "e+00" in x:
        x = x.replace("e+00", "")
    x = x.replace("e+", "e")
    return x


def _dict_job(key, f):
    r = f[0](*f[1], **f[2])
    return key, r


def parallel(
    jobs: list[tuple[Callable, tuple, dict]],
    n_jobs: int = None,
    pbar: str = None,
    pbar_position: int = 0,
    return_as: str = None,
) -> list[Any] | Generator[Any, None, None] | dict[Any, Any]:
    """
    Parallelizes a list of jobs.

    Parameters
    ----------
    jobs : list[tuple[Callable, tuple, dict]]
        The jobs to parallelize. The first element of each tuple is a function, the
        second is a tuple of arguments, and the third is a dictionary of keyword
        arguments.
    n_jobs : int, optional
        The number of jobs to run in parallel. If not provided, the number of parallel
        jobs is set to the number of CPU cores.
    pbar : str, optional
        A label for a progress bar. If not provided, no progress bar is shown.
    pbar_position : int, optional
        The position of the progress bar. If not provided, the progress bar is shown at
        the beginning of the output.
    return_as : Literal["list", "generator", "generator_unordered"], optional
        The type of return value. If not provided, the return value is a list.

    Returns
    -------
    list[Any] | Generator[Any, None, None] | dict[Any, Any]
        The result of the parallelized jobs.
    """
    args = {}
    if return_as is not None:
        args["return_as"] = return_as

    if n_jobs is None:
        n_jobs = N_PARALLEL_PROCESSES

    if isinstance(jobs, dict):
        assert return_as is None, "return_as is not supported for dict jobs"
        result = {
            k: v
            for k, v in parallel(
                [delayed(_dict_job)(k, v) for k, v in jobs.items()],
                pbar=pbar,
                return_as="generator_unordered",
                n_jobs=n_jobs,
                pbar_position=pbar_position,
            )
        }
        return {k: result[k] for k in jobs}

    jobs = list(jobs)

    if n_jobs == 1 or len(jobs) == 1:
        if pbar:
            jobs = tqdm(
                jobs, total=len(jobs), desc=pbar, leave=True, position=pbar_position
            )
        return [j[0](*j[1], **j[2]) for j in jobs]

    total_jobs = len(jobs)

    pbar = tqdm(total=total_jobs, desc=pbar, leave=True) if pbar else None

    def yield_results():
        for result in Parallel(n_jobs=n_jobs, **args)(jobs):
            if pbar:
                pbar.update(1)
            yield result
        if pbar:
            pbar.close()

    if return_as in ["generator", "generator_unordered"]:
        return yield_results()

    # Coerce into generator_unordered so that the progress bar won't hang on one slow
    # job.
    def f(i, job):
        return i, job[0](*job[1], **job[2])

    jobs = [delayed(f)(i, job) for i, job in enumerate(jobs)]
    results = [None] * total_jobs
    args["return_as"] = "generator_unordered"
    for i, result in yield_results():
        results[i] = result
    return results


# import cProfile
# import io
# import pstats


# class ProfilePrint:
#     def __init__(self):
#         self.profiler = cProfile.Profile()

#     def __enter__(self):
#         self.profiler.enable()
#         self.n_jobs = N_PARALLEL_PROCESSES
#         set_n_parallel_jobs(1)
#         return self

#     def __exit__(self, exc_type, exc_value, traceback):
#         self.profiler.disable()
#         s = io.StringIO()
#         stats = pstats.Stats(self.profiler, stream=s).sort_stats("cumulative")
#         stats.print_stats(20)
#         print("\n===== Profiling Results (sorted by total time) =====")
#         print(s.getvalue())
#         # set_n_parallel_jobs(self.n_jobs)


class _SVGJupyterRender(str):
    def _repr_svg_(self):
        return self


def _memmap_read(x):
    f = tempfile.NamedTemporaryFile(delete=False, suffix=".pkl")
    joblib.dump(x, f.name)
    return joblib.load(f.name, mmap_mode="r")



from accelforge.frontend.renames import TensorName


import itertools
from enum import Enum

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import (
    MappingNode,
    Toll,
    Temporal,
    Spatial,
    TensorHolder,
)
from accelforge.frontend.workload import (
    Einsum,
    RankVariable,
    Workload,
)


# =================================================================================================
# Insert loops
# =================================================================================================


class LowerChoice(Enum):
    YES = 0
    NO = 1
    OPTIONAL = 2


def insert_temporal_loops(
    mapping: list[TensorHolder],
    einsum: Einsum,
    first_memory: arch.Memory,
    rank_variable_bounds: dict[RankVariable, int],
    ranks_with_tile_pattern: set,
    workload: Workload,
    _can_lower_outermost_memory: bool,
    flattened_arch: list[arch.Leaf],
    max_fused_loops: int,
    fanouts: dict[str, int],
):
    # First establish insertion points. Insertion points are:
    # - Below the last instance of the first memory
    # - Between any two TensorHolder nodes
    # - After the last TensorHolder node

    # The following logic is really just to make sure that all the storage nodse for the
    # outermost memory are together at the beginning of the split mapping. After that,
    # each entries in the split mapping has a single TensorHolder.
    split_mapping: list[list[TensorHolder]] = [[]]
    for m in mapping:
        split_mapping.append([m])
        if len(split_mapping) > 1 and m.component == first_memory.name:
            split_mapping[-2].extend(split_mapping.pop(-1))
    for i, s in enumerate[list[TensorHolder | Spatial]](split_mapping):
        for m in s:
            if i == 0 and m.component != first_memory.name:
                raise ValueError(
                    "The first TensorHolder in the mapping is not for the outermost "
                    "memory. This isn't known to be invalid, but the code may not "
                    "handle it."
                )
            elif i > 0 and m.component == first_memory.name:
                raise ValueError(
                    "First memory isn't at the top of the hierarchy. This isn't known"
                    "to be invalid, but the code may not handle it."
                )
            elif i == 0 and isinstance(m, Spatial):
                raise ValueError(
                    "Found Spatial node before any TensorHolder. This isn't known to "
                    "be invalid, but the code may not handle it."
                )

    split_mapping = [m for m in split_mapping if m]

    # These Einsum properties are recalculated since Einsum is mutable
    # We're pre-computing and reusing for efficiency
    tensor2fully_relevant_rank_vars = einsum.tensor2directly_indexing_rank_variables
    tensor2partially_relevant_rank_vars = (
        einsum.tensor2expression_indexing_rank_variables
    )
    for k, v in tensor2partially_relevant_rank_vars.items():
        tensor2partially_relevant_rank_vars[k] = v - tensor2fully_relevant_rank_vars[k]
    tensor2irrelevant_rank_vars = einsum.tensor2irrelevant_rank_variables
    tensor2rank_vars = einsum.tensor2rank_variables
    tensors = einsum.tensor_names

    fusable_tensors = (
        einsum.tensor_names & workload.tensor_names_used_in_multiple_einsums
    )
    is_fused_loops = True
    seen_tensors = set()
    choices = []
    lowering_choices: list[tuple[bool, ...]] = []

    def _get_next_storages(i: int, toll_allowed: bool = False) -> list[TensorHolder]:
        for j in range(i + 1, len(split_mapping)):
            assert len(split_mapping[j]) <= 1
            # We don't add loops before tolls since they don't reuse things
            if isinstance(split_mapping[j][0], Toll) and not toll_allowed:
                continue
            return split_mapping[j]
        return []

    prev_fanout = 1
    someone_elses_spatials_may_be_placed_above = False
    for i, prev_storages in enumerate(split_mapping):
        # =============================================================================
        # Choose what temporal loops to insert between prev_storages and the next
        # TensorHolder node(s).
        # =============================================================================

        next_storages = _get_next_storages(i)
        next_anything = _get_next_storages(i, toll_allowed=True)

        for s in prev_storages:
            # No tensor holders must mix backing/non-backing tensors.
            assert not s._backing or all(t in s._backing for t in s.tensors)
            # One tensor per holder
            assert len(s.tensors) == 1

        rank_variables = einsum.rank_variables
        # rank_variables = {r for r in rank_variables if rank_variable_bounds[r] > 1}
        seen_tensors |= set.union(*(set(t.tensors) for t in prev_storages), set())
        is_fused_loops = is_fused_loops and len(fusable_tensors - seen_tensors) > 0
        prev_tensors = set.union(set(), *(set(t.tensors) for t in prev_storages))
        next_persistent = set.union(
            set(), *(set(t.tensors) for t in next_storages if t.persistent)
        )

        max_fanout_before = max(
            [fanouts[s2.component] for s in split_mapping[:i] for s2 in s],
            default=float("inf"),
        )
        min_fanout_after = min(
            [fanouts[s2.component] for s in split_mapping[i + 1 :] for s2 in s],
            default=0,
        )
        cur_fanout = set(fanouts[s2.component] for s2 in prev_storages)
        next_fanout = set(fanouts[s2.component] for s2 in next_anything)
        if len(next_fanout) == 0:
            next_fanout.add(float("inf"))
        # Either it's main memory or we have one entry in the list, so there should only
        # be one
        assert len(cur_fanout) == 1
        assert len(next_fanout) == 1
        cur_fanout = next(iter(cur_fanout))
        next_fanout = next(iter(next_fanout))

        # Can't have loops above persistent tensor holders
        if next_persistent:
            rank_variables &= set()

        # No recomputation: If we haven't seen a tensor yet, must only iterate over
        # fully-relevant rank variables.
        for t in tensors - seen_tensors:
            rank_variables &= tensor2fully_relevant_rank_vars[t]

        if max_fused_loops == 0 and (fusable_tensors - seen_tensors):
            rank_variables &= set()

        #  The fanout for a prior node may be placed here, so spatial nodes may be moved
        #  here
        someone_elses_spatials_may_be_placed_below = (
            next_fanout > cur_fanout and max_fanout_before > cur_fanout
        )

        # If the fanout is about to increase, then spatial loops may be placed below the
        # current node. There may have been constrained temporal loops earlier that need
        # to be placed here, so we won't prohibit any loops.
        if someone_elses_spatials_may_be_placed_below:
            pass
        else:

            # Optimality-preserving optimization: Loops below tolls aren't helpful
            # because there is no storage. Ctrl-F for
            # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put another
            # node's spatial loops below this one, because lowering would add move the
            # spatials down, which would constrain the temporals due to spatial-temporal
            # crossing.
            if isinstance(prev_storages[0], Toll):
                rank_variables &= set()

            # Generally we want to only use rank variables that are irrelevant to the
            # previous tensors, else we'd just lower those tensors. However, we can't
            # lower backing TensorHolder nodes because this will add loops to
            # compatibility.

            # Optimality-preserving optimization: We can trivially lower non-backing
            # TensorHolder nodes through fully-relevant loops. Can't do this if the
            # loops are fused because that'd add loops to the compatibility. Ctrl-F
            # forCONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put
            # another node's spatial loops below this one, because lowering would add
            # move the spatials down, which would constrain the temporals due to
            # spatial-temporal crossing.
            for s in prev_storages:
                for t in s.tensors:
                    if t not in s._backing and not s._must_be_here:
                        rank_variables -= tensor2fully_relevant_rank_vars[t]

            # Optimality-preserving optimization: We can trivially raise TensorHolder
            # nodes through irrelevant unfused loops. Can't do this if the loops are
            # fused because that'd increase the lifetime of the TensorHolder node. Can't
            # do this if the irrelevant rank variables partially-relevant to the
            # previous tensors, since that affects the permutation. See
            # CONTIGUOUS_ITERATION_SPACE_DISCUSSION: Can't do this if we may put another
            # node's spatial loops above this one, because raising would add move the
            # temporals down, which would constrain them due to spatial-temporal
            # crossing. TODO: CONTIGUOUS_ITERATION_SPACE_DISCUSSION: This causes all
            # loops to be added, but really we only need to re-add the ones that may
            # conflict with a spatial loop.
            if not is_fused_loops:
                for s in next_storages:
                    if not s._must_be_here:
                        for t in s.tensors:
                            rvs = tensor2irrelevant_rank_vars[t]
                            for t2 in prev_tensors:
                                rvs -= tensor2partially_relevant_rank_vars[t2]
                            rank_variables -= rvs

        # =============================================================================
        # Determine whether to lower TensorHolder nodes through partially-relevant
        # loops.
        # =============================================================================
        partially_relevant_to_previous = rank_variables & set.union(
            set(), *(tensor2partially_relevant_rank_vars[t] for t in prev_tensors)
        )
        permutable_partially_relevant = set()

        # NOTE: If the lowering logic for backing TensorHolders is updated & we can
        # lower through >1 loops, then also update label_fused_loops
        for s in prev_storages:
            partially_relevant_to_previous = set.union(
                set(), *(tensor2partially_relevant_rank_vars[t] for t in s.tensors)
            )
            partially_relevant_to_previous &= rank_variables
            lowerable_backing = (
                _can_lower_outermost_memory or s.component != first_memory.name
            )

            # Persistent. Must be at the top of the mapping.
            if s.persistent:
                lowering_choices.append((False,))
            # Don't lower our own reservations through someone else's spatial loops.
            elif someone_elses_spatials_may_be_placed_below:
                lowering_choices.append((False,))
            # Processing stage. Lowering doesn't matter. Don't lower.
            elif isinstance(s, Toll):
                lowering_choices.append((False,))
            # Previous is backing and there's partially-relevant rank variables. May
            # want to lower to reduce memory footprint, or raise to reduce number of
            # fused loops.
            elif s._backing and lowerable_backing and partially_relevant_to_previous:
                assert False
                lowering_choices.append((False, True))
                permutable_partially_relevant |= partially_relevant_to_previous
            # No backing in previous. No cost to lowering. Lower all
            elif not s._backing:
                # print(f'Can lower {s.tensors} in {s.component}')
                lowering_choices.append((True,))
                permutable_partially_relevant |= partially_relevant_to_previous
            # Previous TensorHolder is backing but not lowerable or there are no
            # partially relevant rank vars.
            else:
                lowering_choices.append((False,))

        # =============================================================================
        # Create loop order and lowering choices
        # =============================================================================

        can_lower = any(any(c) for c in lowering_choices)

        # Create canonical loop orders that avoids repeating reuse patterns.
        choices.append(
            list(
                canonical_loop_orders(
                    rank_variables, permutable_partially_relevant, can_lower
                )
            )
        )
        prev_fanout = cur_fanout
        someone_elses_spatials_may_be_placed_above = (
            someone_elses_spatials_may_be_placed_below
        )

    # ==================================================================================
    # Iterate over all possible mappings
    # ==================================================================================

    # TODO: Optimization: If we can optionally lower a tensor & the loop below it is
    # not something through which we can lower for a given permutation, skip options
    # that lower that tensor because they get the same result as not lowering the
    # tensor.
    n_loop_orders = len(list(itertools.product(*choices)))
    for loop_orders in itertools.product(*choices):
        full_mapping = []
        for prev_storages, loop_order in zip(split_mapping, loop_orders):
            full_mapping.extend(prev_storages)
            full_mapping.extend(Temporal(rank_variable=r) for r in loop_order)

        storages = [node for node in full_mapping if isinstance(node, TensorHolder)]
        assert len(lowering_choices) == len(storages)
        for lowering_choice in itertools.product(*lowering_choices):
            for lower, node in zip(lowering_choice, storages):
                node._lower = lower

            yield list(full_mapping), n_loop_orders


def insert_spatial_loops(
    mapping: list[MappingNode],
    einsum: Einsum,
    flattened_arch: list[arch.Memory],
):
    nodes_with_fanout = [n for n in flattened_arch if n.get_fanout() > 1]
    arch_node_names = [n.name for n in flattened_arch]
    tensors = einsum.tensor_names
    tensor2fully_relevant_rank_vars = einsum.tensor2directly_indexing_rank_variables

    for node in nodes_with_fanout:
        insertion_point = _idx_of_highest_tensor_holder_with_component_below_fanout(
            node, mapping, arch_node_names
        )

        # No recomputation: If we haven't seen a tensor yet, must only iterate over
        # fully-relevant rank variables.
        rank_variables = einsum.rank_variables
        for t in tensors - _tensors_seen_above_point(insertion_point, mapping):
            rank_variables &= tensor2fully_relevant_rank_vars[t]

        for fanout_dim in node.spatial:
            for r in sorted(rank_variables):
                s = Spatial(
                    rank_variable=r,
                    name=fanout_dim.name,
                    component_object=node,
                    component=node.name,
                )
                if insertion_point == len(mapping):
                    mapping.append(s)
                else:
                    mapping.insert(insertion_point, s)


def _tensors_seen_above_point(idx, mapping):
    seen_tensors = set()
    for i in range(idx):
        node = mapping[i]
        if not isinstance(node, TensorHolder):
            continue
        seen_tensors |= set(node.tensors)
    return seen_tensors


def _idx_of_highest_tensor_holder_with_component_below_fanout(
    fanout_node, mapping, arch_node_names
):
    for i in range(len(mapping)):
        if not isinstance(mapping[i], TensorHolder):
            continue
        if arch_node_names.index(mapping[i].component) >= arch_node_names.index(
            fanout_node.name
        ):
            return i
    return len(mapping)


def canonical_loop_orders(
    rank_variables: set[RankVariable],
    partially_relevant_to_previous: set[RankVariable],
    can_lower: bool,
):
    """Generate loop orders that result in unique reuse patterns."""
    # Only the first partially-relevant rank variable matters is a meaningful
    # choice because lowering only happens through at most one rank var.
    if not partially_relevant_to_previous or not can_lower:
        yield tuple(sorted(rank_variables))
        return

    for first_rank_var in partially_relevant_to_previous:
        rest_of_partially_relevant = partially_relevant_to_previous - {first_rank_var}
        rest_rank_vars = rank_variables - partially_relevant_to_previous
        # Since order does not matter, we choose alphabetical order as canonical.
        yield (
            (first_rank_var,)
            + tuple(sorted(rest_of_partially_relevant))
            + tuple(sorted(rest_rank_vars))
        )



import copy
from collections.abc import Generator
from itertools import chain, combinations
import logging

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import Storage, TensorHolder, Toll
from accelforge.frontend.workload import TensorName, SymbolTable

from accelforge.util.exceptions import EvaluationError
from accelforge.util._setexpressions import InvertibleSet


def make_tensor_choices_one_level(
    node: arch.Leaf,
    symbol_table: dict[str, InvertibleSet],
    persistent_tensors: set[TensorName],
    seen_tensors: set[TensorName] = (),
    is_copy_op: bool = False,
    einsum_name: str = None,
) -> Generator[tuple[list[TensorHolder], SymbolTable, set[TensorName]], None, None]:
    """
    Generate combinations of TensorHolder nodes based on keep and bypass
    constraints.

    Each generated list contains TensorHolder nodes for single tensors.
    """
    assert "All" in symbol_table
    tensors = symbol_table["All"]

    if not isinstance(node, arch.TensorHolder):
        yield [], symbol_table, set(seen_tensors)
        return

    if isinstance(node, arch.Memory):
        target_type = Storage
    elif isinstance(node, arch.Toll):
        target_type = Toll
    elif isinstance(node, arch.Dummy):
        yield [], symbol_table, set(seen_tensors)
        return
    else:
        raise ValueError(f"Unexpected tensor holder type: {type(node)}")

    new_symbol_table = copy.copy(symbol_table)

    node = copy.copy(node)
    try:
        node.tensors: arch.Tensors = node.tensors._eval_expressions(
            symbol_table=symbol_table,
            musteval_tryeval_to=True,
            must_copy=False,
            location=f"arch.{node.name}.tensors",
        )[0]
    except EvaluationError as e:
        e.add_field(f"Einsum {einsum_name} arch.{node.name}.tensors")
        raise e

    must_keep = tensors.to_my_space(node.tensors.keep | node.tensors.back)
    may_keep = tensors.to_my_space(node.tensors.may_keep)
    may_keep -= must_keep

    if seen_tensors & set(node.tensors.back):
        return

    if must_keep - tensors:
        raise KeyError(
            f"Keep constraint for {node.name} includes tensors that are "
            f"not in the workload: {must_keep - new_symbol_table['All']}"
        )
    if may_keep - tensors:
        raise KeyError(
            f"Bypass constraint for {node.name} includes tensors that are "
            f"not in the workload: {may_keep - tensors.full_space}"
        )

    logging.info(
        f"\t\t{node.name} must keep {sorted(must_keep)}, may keep {sorted(may_keep)}"
    )

    # No reuse in copy operations, so no need to keep tensors in more places
    if is_copy_op:
        may_keep -= tensors.to_my_space(seen_tensors)

    for subset in powerset(sorted(may_keep, key=str)):
        # Make keep choice & update symbol table
        subset = tensors.to_my_space(set(subset))
        keep_choice = tensors.to_my_space(subset | must_keep)
        # Below line is so users can do MainMemory().tensors() or MainMemory.tensors
        new_symbol_table[node.name] = keep_choice
        new_symbol_table["Above"] |= keep_choice
        new_seen_tensors = seen_tensors | set(keep_choice)

        # Make sure they're all tensors
        assert all(isinstance(k, TensorName) for k in keep_choice)
        keep_choice = keep_choice.to_my_space({copy.copy(t) for t in keep_choice})
        nodes = []

        # Create storage nodes. Sort them to keep this deterministic. Ordering is done
        # later.
        for t in sorted(keep_choice, key=str):
            nodes.append(
                target_type(tensors=[t], component=node.name, component_object=node)
            )
            if t not in seen_tensors:
                nodes[-1]._backing.add(t)
                nodes[-1]._must_keep_tensors = [t]
                nodes[-1].persistent = t in persistent_tensors
            elif t in must_keep:
                nodes[-1]._must_keep_tensors = [t]

        yield nodes, new_symbol_table, new_seen_tensors


def make_storage_choices_all_levels(
    nodes: list[TensorHolder],
    symbol_table: dict[str, InvertibleSet],
    persistent_tensors: set[TensorName],
    seen_tensors: set[TensorName] = None,
    is_copy_op: bool = False,
    einsum_name: str = None,
) -> Generator[tuple[dict[str, list[TensorHolder]], SymbolTable], None, None]:
    """
    Generate combinations of TensorHolder nodes based on keep and bypass
    constraints.

    Each generated dict maps memory name to a list of TensorHolder nodes for
    single tensors.
    """
    seen_tensors = set() if seen_tensors is None else seen_tensors
    if len(nodes) == 0:
        yield dict(), symbol_table
        return

    if "Above" not in symbol_table:
        symbol_table["Above"] = symbol_table["Nothing"]

    for choice, symbol_table, new_seen_tensors in make_tensor_choices_one_level(
        node=nodes[0],
        symbol_table=symbol_table,
        persistent_tensors=persistent_tensors,
        seen_tensors=seen_tensors,
        is_copy_op=is_copy_op,
        einsum_name=einsum_name,
    ):
        for subchoices, symbol_table in make_storage_choices_all_levels(
            nodes=nodes[1:],
            symbol_table=symbol_table,
            persistent_tensors=persistent_tensors,
            seen_tensors=new_seen_tensors,
            is_copy_op=is_copy_op,
            einsum_name=einsum_name,
        ):
            yield {**subchoices, nodes[0].name: choice}, symbol_table


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))



from .mapping import *

__all__ = [
    "Compute",
    "Loop",
    "Mapping",
    "MappingNode",
    "MappingNodeWithChildren",
    "Nested",
    "NodeList",
    "Pipeline",
    "RankVariable",
    "Reservation",
    "Sequential",
    "Spatial",
    "Split",
    "Storage",
    "Temporal",
    "TensorHolder",
    "TensorName",
    "TilePattern",
    "Toll",
]



from accelforge.frontend.mapping import (
    Mapping,
)

from accelforge.frontend.mapping import Mapping, TensorHolder, Compute as ComputeSpec
from accelforge.frontend.workload import (
    TensorName,
)
from accelforge.model._looptree.types import Buffet, Compute


class DataMovementConnections:
    """
    For querying connection between Buffet or Compute
    (e.g., which Buffet supplies this Compute).
    """

    def __init__(self):
        self.src_to_dst = {}
        self.dst_to_src = {}

    def get_src(self, dst: Buffet | Compute) -> Buffet | None:
        return self.dst_to_src[dst]

    def get_dst(self, src: TensorHolder) -> Buffet | Compute:
        return self.src_to_dst[src]

    @property
    def sources(self) -> list[TensorHolder]:
        return list(self.src_to_dst.keys())

    @property
    def destinations(self) -> list[TensorHolder]:
        return list(self.dst_to_src.keys())

    @classmethod
    def from_pmapping(cls, pmapping):
        einsum_name = pmapping[-1].einsum
        src_to_dst = {}
        dst_to_src = {}
        tensor_to_last_src = {}
        for node in pmapping:
            if isinstance(node, TensorHolder):
                for tensor in node.tensors:
                    buffet = Buffet(tensor, einsum_name, node.component)
                    if tensor in tensor_to_last_src:
                        src = tensor_to_last_src[tensor]
                    else:
                        src = None
                    tensor_to_last_src[tensor] = buffet
                    dst_to_src[buffet] = src
                    if src is not None:
                        src_to_dst[src] = buffet
                    src_to_dst[buffet] = None
            elif isinstance(node, ComputeSpec):
                for tensor in tensor_to_last_src:
                    dst_buffet = Buffet(tensor, einsum_name, node.component)
                    src = tensor_to_last_src[tensor]
                    src_to_dst[src] = dst_buffet
                    dst_to_src[dst_buffet] = src

        result = DataMovementConnections()
        result.src_to_dst = src_to_dst
        result.dst_to_src = dst_to_src
        return result


def get_tensor_to_backer_id(mapping: Mapping):
    tensor_to_ids: dict[TensorName, set[int]] = {}
    for node in mapping:
        if isinstance(node, TensorHolder):
            for tensor in node.tensors:
                if tensor in tensor_to_ids:
                    continue
                tensor_to_ids[tensor] = id(node)
    return tensor_to_ids



from accelforge.frontend.renames import EinsumName
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup
import inspect
import os
import random
from typing import Callable, Generator
from accelforge import arch, util
from accelforge import Spec
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.mapper.FFM.pmappings import MultiEinsumPmappings
from accelforge.mapper.FFM._join_pmappings.compress_pmappings import (
    compress_einsum2pmappings,
    decompress_pmappings,
)
from accelforge.frontend.workload import EinsumName
from accelforge.frontend.mapping import Mapping
from accelforge.mapper.FFM import PmappingGroup
from accelforge.mapper.FFM._pareto_df.df_convention import (
    MAPPING_COLUMN,
    col2nameloop,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingDataframe
from accelforge.mapper.FFM._make_pmappings.make_pmappings import (
    get_rank_variable_bounds_for_all_einsums,
)
from accelforge._accelerated_imports import pd
import joblib
from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
from accelforge.mapper._simanneal2.tracking import EvaluationsScoreTracker
from accelforge.util.parallel import get_n_parallel_jobs

# Simulated annealing algorithm
# -----------------------------
# Given:
# - Pmappings for each Einsum

# 1. Make a compatibility -> PmappingGroups dict for each Einsum
# 2. While True:
#    a. Randomly change a compatibility choice for one Einsum


# Functions:
# - Given compatibility choices & pmapping index numbers, return a score
# - Given compatibility choices & pmapping index numbers, make sure all compatibilities
#   & indices match


class FailedMutation(Exception):
    pass


class MapspaceGlobals:
    def __init__(
        self,
        einsum2sims: dict[EinsumName, list[PmappingGroup]],
        aliased_tensors: dict[str, set[str]],
        objective_function: Callable[[pd.Series], float],
        tracker: EvaluationsScoreTracker,
    ) -> None:
        self.einsum2sims: dict[EinsumName, list[PmappingGroup]] = einsum2sims
        self.aliased_tensors: dict[str, set[str]] = aliased_tensors
        self.objective_function: Callable[[pd.Series], float] = objective_function
        self.tracker: EvaluationsScoreTracker = tracker


class SimAnnealMapping:
    def __init__(self, mapspace_globals: MapspaceGlobals) -> None:
        # self.einsum2sim: dict[EinsumName, PmappingGroup] = {
        #     e: random.choice(s) for e, s in mapspace_globals.einsum2sims.items()
        # }
        self.mapspace_globals: MapspaceGlobals = mapspace_globals
        self.einsum2sim: dict[EinsumName, PmappingGroup] = {
            e: random.choice(s) for e, s in mapspace_globals.einsum2sims.items()
        }
        self.einsum2index: dict[EinsumName, int] = {e: 0 for e in self.einsum2sim}
        self.init_choices()
        self._prev_score = None

    def mutate(self) -> None:
        # Pick a random einsum
        e = random.choice(list(self.einsum2sim.keys()))

        random.choice(
            [
                self._randomize_index,
                self._randomize_sim,
            ]
        )(e)
        try:
            self.ensure_match(e)
        except FailedMutation:
            raise
        finally:
            self.add_evaluations()

    def add_evaluations(self) -> None:
        s = sum(sim.mappings.n_total_pmappings for sim in self.einsum2sim.values())
        s *= len(self.mapspace_globals.einsum2sims) / len(self.einsum2sim)
        self.mapspace_globals.tracker.add_evaluation(s, float("inf"))

    def _randomize_index(self, e: EinsumName) -> None:
        self._prev_score = None
        self.einsum2index[e] = random.randint(0, 10000000000000)

    def _randomize_sim(self, e: EinsumName) -> None:
        self.einsum2sim[e] = random.choice(self.mapspace_globals.einsum2sims[e])
        self._randomize_index(e)

    def _einsum_position_in_list(self, e: EinsumName) -> int:
        return list(self.einsum2sim.keys()).index(e)

    def init_choices(self) -> None:
        try:
            self.einsum2sim = {}
            for einsum_name in self.mapspace_globals.einsum2sims:
                possible_choices = self.mapspace_globals.einsum2sims[einsum_name]
                cur_tensors = self._einsum2tensors(einsum_name)

                for prev_einsum, prev_sim in self.einsum2sim.items():
                    prev_tensors = self._einsum2tensors(prev_einsum)
                    shared_tensors = prev_tensors & cur_tensors
                    c = prev_sim.compatibility.clear_dead_tensors(
                        shared_tensors
                    ).clear_tile_patterns_and_reservation_indices()

                    possible_choices = [
                        s
                        for s in possible_choices
                        if s.compatibility.clear_dead_tensors(
                            shared_tensors
                        ).clear_tile_patterns_and_reservation_indices()
                        == c
                    ]
                if not possible_choices:

                    raise FailedMutation(
                        f"No compatible PmappingGroups found for {einsum_name}"
                    )
                self.einsum2sim[einsum_name] = random.choice(possible_choices)
                self._randomize_index(einsum_name)
        except FailedMutation:
            raise
        finally:
            self.add_evaluations()

    def ensure_match(
        self,
        lock_choice_for_einsum: EinsumName,
    ) -> None:

        new_einsum2sim: dict[EinsumName, PmappingGroup] = {}

        # Grab all the compatibilities that match
        for i, (e, s) in enumerate[tuple[EinsumName, PmappingGroup]](
            list(self.einsum2sim.items())
        ):
            if e == lock_choice_for_einsum:
                new_einsum2sim[e] = s
                continue

            following_tensors = self._einsum2tensors(range(i + 1, len(self.einsum2sim)))

            to_check = [(s2, s) for s2 in new_einsum2sim.values()]

            if i < self._einsum_position_in_list(lock_choice_for_einsum):
                to_check.append((s, self.einsum2sim[lock_choice_for_einsum]))
            else:
                to_check.append((self.einsum2sim[lock_choice_for_einsum], s))

            for left, right in to_check:
                c = left.compatibility.clear_dead_tensors(
                    right.compatibility.tensor_names
                ).clear_tile_patterns_and_reservation_indices()
                c2 = right.compatibility.clear_dead_tensors(
                    left.compatibility.tensor_names
                ).clear_tile_patterns_and_reservation_indices()
                if c != c2:
                    break

                c = left.compatibility.clear_dead_tensors(
                    following_tensors
                ).clear_tile_patterns_and_reservation_indices()
                c2 = right.compatibility.clear_dead_tensors(
                    following_tensors
                ).clear_tile_patterns_and_reservation_indices()

                # Can't merge. I have more loops than the next, so my dataflow can't be
                # carried through a LoopTree to where it's needed.
                if c.n_loops > c2.n_loops:
                    break

            else:
                new_einsum2sim[e] = s

        # Grab compatibilities that don't match
        def _matches(s: PmappingGroup, c: Compatibility) -> bool:
            cs = s.compatibility.clear_dead_tensors(
                c.tensor_names
            ).clear_tile_patterns_and_reservation_indices()
            cn = c.clear_dead_tensors(
                s.compatibility.tensor_names
            ).clear_tile_patterns_and_reservation_indices()
            return cs == cn

        for e, pmapping_groups in self.mapspace_globals.einsum2sims.items():
            if e in new_einsum2sim:
                continue

            for s in new_einsum2sim.values():
                pmapping_groups = [
                    s2 for s2 in pmapping_groups if _matches(s2, s.compatibility)
                ]

            if not pmapping_groups:
                # print(f"No compatible PmappingGroups found for {e}")
                raise FailedMutation(f"No compatible PmappingGroups found for {e}")

            new_einsum2sim[e] = random.choice(pmapping_groups)
            self._randomize_index(e)

            # pmapping_groups = self.mapspace_globals.einsum2sims[e]
            # [s.compatibility for s in self.einsum2sim.values()]
            # [s.compatibility for s in new_einsum2sim.values()]
            # {e: s.compatibility for e, s in new_einsum2sim.items()}

        assert len(new_einsum2sim) == len(self.einsum2sim)
        assert set(new_einsum2sim.keys()) == set(self.einsum2sim.keys())
        self.einsum2sim = {k: new_einsum2sim[k] for k in self.einsum2sim.keys()}

    def _einsum2tensors(
        self, e: EinsumName | int | Generator[EinsumName | int, None, None]
    ) -> set[str]:
        if isinstance(e, Generator) or isinstance(e, range):
            return set.union(set(), *(self._einsum2tensors(i) for i in e))
        if isinstance(e, int):
            e = list(self.mapspace_globals.einsum2sims.keys())[e]
        return self.mapspace_globals.einsum2sims[e][0].compatibility.tensor_names

    def _access_index(self, e: EinsumName, index_override: int | None = None):
        s = self.einsum2sim[e]
        data = s.mappings.data
        i = self.einsum2index[e] if index_override is None else index_override
        i %= len(data)
        return PmappingGroup(
            compatibility=s.compatibility,
            mappings=PmappingDataframe(
                data.iloc[i : i + 1],
                n_total_pmappings=s.mappings.n_total_pmappings,
                n_valid_pmappings=s.mappings.n_valid_pmappings,
            ),
        )

    def get_score(self) -> float:
        if self._prev_score is not None:
            return self._prev_score

        items: list[tuple[EinsumName, PmappingGroup]] = list(self.einsum2sim.items())
        joined: PmappingGroup = items.pop(0)[1]
        for i, (e, s) in enumerate(items):
            right_tensors = self._einsum2tensors(i)
            live_tensors = self._einsum2tensors(range(i + 1, len(items)))

            joined.compatibility = joined.compatibility.clear_dead_tensors(
                live_tensors | right_tensors
            )

            def _merge_next(
                left: PmappingGroup,
                right: PmappingGroup,
            ) -> PmappingGroup:
                try:
                    return left.merge_next(
                        right,
                        live_tensors=live_tensors,
                        live_tensors_with_right=live_tensors | right_tensors,
                        aliased_tensors=self.mapspace_globals.aliased_tensors,
                        compatibility_joined=joined.compatibility.merge_next(
                            s.compatibility,
                            live_tensors,
                        ),
                        permuted_compatibility_left=left.compatibility,
                        permuted_compatibility_right=right.compatibility,
                        drop_valid_reservations=True,
                        delay=False,
                        ignored_resources=set(),
                    )
                except ValueError as err:
                    # print(err)
                    raise FailedMutation(f"No valid pmappings: {err}")

            # Try to merge using the index we already have set
            joined_new = _merge_next(joined, self._access_index(e))
            if len(joined_new.mappings.data) == 1:
                joined = joined_new
                # print(' '.join(f'{k}={v}' for k, v in dict(joined.mappings.data.iloc[0]).items() if col2nameloop(k)))
                continue
            if len(joined_new.mappings.data) > 1:
                raise ValueError(
                    f"Got {len(joined_new.mappings.data)} pmappings for {e}"
                )

            # No valid pmappings! Merge all possible, then pick one. We'll charge for
            # evaluations again because we're picking a new mapping.

            self.add_evaluations()

            s = self.einsum2sim[e]
            s.mappings.data["_INDEX"] = list(range(len(s.mappings.data)))
            joined_new = _merge_next(
                joined,
                s,
            )
            s.mappings._data = s.mappings.data.drop(columns=["_INDEX"])
            try:
                i = random.choice(list(set(joined_new.mappings.data["_INDEX"])))
            except IndexError:
                raise FailedMutation(f"No valid pmappings for {e}")

            # Now that we've picked, merge with the index we just set
            joined_new = _merge_next(joined, self._access_index(e, i))

            if len(joined_new.mappings.data) == 1:
                # If it worked, set the index
                self.einsum2index[e] = i
                joined = joined_new
                # print(' '.join(f'{k}={v}' for k, v in dict(joined.mappings.data.iloc[0]).items() if col2nameloop(k)))
                continue

            if len(joined_new.mappings.data) > 1:
                raise ValueError(
                    f"Got {len(joined_new.mappings.data)} pmappings for {e}"
                )

            raise FailedMutation(
                f"Got {len(joined_new.mappings.data)} pmappings for {e}"
            )

        assert len(joined.mappings.data) == 1
        score = self.mapspace_globals.objective_function(joined.mappings.data.iloc[0])
        self.mapspace_globals.tracker.add_evaluation(0, score)
        self._prev_score = score
        return score

    def crossover(self, other: "SimAnnealMapping") -> "SimAnnealMapping":
        child = other.copy()
        e = random.choice(list(self.einsum2sim.keys()))
        child.einsum2sim[e] = self.einsum2sim[e]
        child.einsum2index[e] = self.einsum2index[e]

        try:
            child.ensure_match(e)
        except FailedMutation:
            child = other.copy()

        child.add_evaluations()
        return child

    def copy(self) -> "SimAnnealMapping":
        s = SimAnnealMapping.__new__(SimAnnealMapping)
        s.mapspace_globals = self.mapspace_globals
        s.einsum2sim = self.einsum2sim.copy()
        s.einsum2index = self.einsum2index.copy()
        s._prev_score = self._prev_score
        return s


def get_random_mapping(mapspace_globals: MapspaceGlobals) -> SimAnnealMapping:
    while True:
        try:
            s = SimAnnealMapping(mapspace_globals)
            s.get_score()
            return s
        except FailedMutation:
            if mapspace_globals.tracker.finished():
                return None
            continue


def _make_mapspace_globals(
    pmapping_groups: dict[EinsumName, list[PmappingGroup]],
    spec: Spec,
    tracker: EvaluationsScoreTracker,
) -> MapspaceGlobals:
    objective = spec.mapper.metrics
    if objective == Metrics.ENERGY:
        objective_function = lambda x: x["Total<SEP>energy"]
    elif objective == Metrics.LATENCY:
        objective_function = lambda x: x["Total<SEP>latency"]
    elif objective == (Metrics.ENERGY | Metrics.LATENCY):
        objective_function = lambda x: x["Total<SEP>energy"] * x["Total<SEP>latency"]
    else:
        raise ValueError(f"Unknown objective {objective}")
    return MapspaceGlobals(
        einsum2sims=pmapping_groups,
        aliased_tensors=spec.workload.get_tensor_copies(),
        objective_function=objective_function,
        tracker=tracker,
    )


def _make_initial_population(
    mapspace_globals: MapspaceGlobals,
    pop_size: int,
) -> list[SimAnnealMapping]:
    tracker = mapspace_globals.tracker
    mappings = []
    while len(mappings) < pop_size:
        m = get_random_mapping(mapspace_globals)
        if m is None:
            return mappings
        mappings.append(m)
        if tracker.finished():
            print(f"Timed out after {len(mappings)}/{pop_size} mappings.")
            return mappings
    print(f"Completed making initial population of {len(mappings)} mappings")
    return mappings


def _join_pmappings(
    pmapping_groups: dict[EinsumName, list[PmappingGroup]],
    spec: Spec,
    tracker: EvaluationsScoreTracker,
    pop_size_per_thread: int,
    algorithm: str = "simanneal",
) -> PmappingGroup:
    mapspace_globals = _make_mapspace_globals(pmapping_groups, spec, tracker)
    mappings = _make_initial_population(mapspace_globals, pop_size_per_thread)
    if not mappings or tracker.finished():
        return

    if algorithm == "simanneal":
        _run_simanneal(mappings, tracker)
    elif algorithm == "genetic":
        _run_genetic(mappings, tracker, pop_size_per_thread)
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")


def _run_simanneal(
    mappings: list[SimAnnealMapping],
    tracker: EvaluationsScoreTracker,
) -> None:
    while not tracker.finished():
        for i, m in enumerate(list(mappings)):
            try:
                new = m.copy()
                new.mutate()
                if new.get_score() < m.get_score():
                    mappings[i] = new
                if tracker.finished():
                    break
            except FailedMutation:
                continue


def _run_genetic(
    mappings: list[SimAnnealMapping],
    tracker: EvaluationsScoreTracker,
    pop_size: int,
    crossover_rate: float = 0.7,
    mutation_rate: float = 0.3,
) -> None:
    while not tracker.finished():
        fitness = []
        for m in mappings:
            try:
                fitness.append(m.get_score())
            except FailedMutation:
                fitness.append(float("inf"))

        best_idx = min(range(len(fitness)), key=lambda i: fitness[i])
        best_mapping = mappings[best_idx]

        inv_fitness = [1.0 / (f + 1e-9) for f in fitness]
        total = sum(inv_fitness)
        probs = [f / total for f in inv_fitness]
        selected = random.choices(range(len(mappings)), probs, k=pop_size)

        new_pop: list[SimAnnealMapping] = []
        for i in range(0, len(selected), 2):
            p1 = mappings[selected[i]]
            p2 = mappings[selected[(i + 1) % len(selected)]]
            if random.random() < crossover_rate:
                new_pop.append(p1.crossover(p2))
                new_pop.append(p2.crossover(p1))
            else:
                new_pop.append(p1.copy())
                new_pop.append(p2.copy())

        for m in new_pop:
            if random.random() < mutation_rate:
                try:
                    m.mutate()
                except FailedMutation:
                    pass

        new_pop.append(best_mapping)
        mappings = new_pop[:pop_size]

        if tracker.finished():
            break


def get_n_tile_shapes(sim: PmappingGroup) -> int:
    df = sim.mappings.data
    symbols = sim.compatibility.symbols()
    cols = [c for c in df.columns if c in symbols]
    if not cols:
        return 1
    return len(df.groupby(cols).size())


def join_pmappings(
    pmappings: MultiEinsumPmappings,
    max_evaluations: int = 1,
    population_size=1000,
    score_target: float | None = None,
    algorithm: str = "simanneal",
) -> EvaluationsScoreTracker:

    spec = pmappings.spec
    tracker = EvaluationsScoreTracker(
        max_evaluations=max_evaluations / get_n_parallel_jobs(),
        stop_at_score=None,
        print_period=1,
    )

    # Disable validation in the compatibility class to avoid errors when joining
    # pmappings. We use a simplified version of the joining process in this code that
    # doesn't do all the reservation cleaning that we do in FFM.
    Compatibility.__post_init__ = lambda *args, **kwargs: None

    compressed, decompress_data = compress_einsum2pmappings(pmappings.einsum2pmappings)

    if score_target is not None:
        tracker._scale_score_by *= 1 / score_target

    pop_size_per_thread = max(1, population_size // get_n_parallel_jobs())

    # Multiply by the number of einsums
    # print(
    #     f"Multiplying scale by {len(pmappings.einsum2pmappings)} for number of einsums"
    # )
    # tracker._scale_by *= len(pmappings.einsum2pmappings)

    # We allow FFM to query n_pareto_optimal_pmappings mappings from the mapspace, so we
    # scale by 1 / n_pareto_optimal_pmappings to get simanneal 1 evaluation = same
    # number of mappings as FFM
    # print(
    #     f"Multiplying scale by {1 / pmappings.n_pareto_optimal_pmappings()} for Pareto-optimal mapspace size"
    # )
    # tracker._scale_by *= 1 / pmappings.n_pareto_optimal_pmappings()

    for einsum_name, einsum_pmappings in pmappings.einsum2pmappings.items():
        total = sum(len(p.mappings.data) for p in einsum_pmappings)
        n_compatibilities = len(einsum_pmappings)
        print(
            f"Einsum {einsum_name} has {total} pmappings with {n_compatibilities} compatibilities"
        )
        if total == 0:
            raise ValueError(f"Einsum {einsum_name} has no pmappings")

    permuted = {}
    n_evaluated = 1
    for einsum_name, einsum_sims in compressed.items():
        for s in einsum_sims:
            n_pmappings = s.mappings.n_total_pmappings
            n_evaluated += n_pmappings
            # Count all permutations as separate choices
            for c_perm, _ in s.compatibility.make_equivalent_permutations():
                permuted.setdefault(einsum_name, []).append(
                    PmappingGroup(
                        compatibility=c_perm,
                        mappings=s.mappings,
                    )
                )

    print(f"Multiplying scale by {1 / n_evaluated} for number of evaluated pmappings")
    tracker._scale_by *= 1 / n_evaluated

    def parallel_join(
        permuted: dict[EinsumName, list[PmappingGroup]],
        spec: Spec,
        tracker: EvaluationsScoreTracker,
        pop_size_per_thread: int,
        algorithm: str,
    ) -> EvaluationsScoreTracker:
        _join_pmappings(permuted, spec, tracker, pop_size_per_thread, algorithm)
        return tracker

    trackers = util.parallel(
        joblib.delayed(parallel_join)(
            permuted,
            spec,
            tracker,
            pop_size_per_thread,
            algorithm,
        )
        for _ in range(get_n_parallel_jobs())
    )

    t0 = trackers[0]
    for t in trackers[1:]:
        t0.merge_with(t)

    # for einsum_name in pmappings.einsum2pmappings:
    #     col = f"{einsum_name}<SEP>{MAPPING_COLUMN}"
    #     joined.data[col] = joined.data[col].apply(
    #         lambda x: pmappings.pmapping_objects[einsum_name][x]
    #     )

    # rank_variable_bounds = get_rank_variable_bounds_for_all_einsums(spec)
    # joined.data[f"Total<SEP>{MAPPING_COLUMN}"] = joined.data.apply(
    #     lambda row: MappingFromRow(row, spec, rank_variable_bounds), axis=1
    # )
    # # Fill nans with 0. We might get missing columns for some mapping entries if there
    # # are energy entries for some pmappings but not others (e.g., one pmapping accesses
    # # DRAM while another doesn't.)
    # joined._data = joined.data.fillna(0)
    return t0  # Mappings(spec, list(pmappings.einsum2pmappings.keys()), joined.data)



from collections import defaultdict
import itertools
import time
from accelforge._accelerated_imports import pd
from accelforge.mapper.FFM._join_pmappings.sim import PmappingGroup, Loop, Compatibility
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingDataframe
from accelforge.mapper.simanneal.mapspaceglobals import MapspaceGlobals
from accelforge.util._frozenset import fzs


def mapping2sims(einsum_to_result: Compatibility):
    r = {}
    for einsum_name, compat_dict in einsum_to_result.items():
        r[einsum_name] = [paretofy(k, v) for k, v in compat_dict.items()]
    return list(r.values())


def get_possible_translations(
    t: Compatibility,
    pairwise_equivalent_rank_variables: dict[str, set[str]],
    full_equivalent_rank_variables: dict[str, set[str]],
    right_rank_variables: set[str],
):
    # Fused ranks should be transitive, but if a fused loop indexes into two
    # different ranks in the next Einsum, we can't fuse becuase it will tile in
    # multiple directions.
    #
    # The first union checks what loops we CAN fuse with in the next Einsum. The
    # second union checks what loops MUST index into in the next
    #
    # Einsum. If we alias into multiple ranks, we can't fuse. Otherwise, try out
    # each possible rank.
    def translate_loop(l: Loop):
        compatible_rank_variables = (
            set.union(
                *(full_equivalent_rank_variables[n] for n in l.rank_variable_names)
            )
            & right_rank_variables
        )
        pairwise_compatible_rank_variables = (
            set.union(
                *(pairwise_equivalent_rank_variables[n] for n in l.rank_variable_names)
            )
            & right_rank_variables
        )
        if len(pairwise_compatible_rank_variables) > 1:
            return
        for n in compatible_rank_variables:
            yield Loop(fzs((n,)), l.bound, l.is_spatial)

    for loops in itertools.product(*map(translate_loop, t.loops)):
        yield t.update(loops=loops)


prev_time = 0
total_time = defaultdict(int)


def init_print_time():
    global prev_time, total_time
    prev_time = time.time()
    total_time = defaultdict(int)


def print_time(what: str):
    global prev_time
    t = time.time() - prev_time
    print(f"{what}: {t:.2f} seconds")
    total_time[what] += t
    prev_time = time.time()


def print_total_time():
    print(f"\n======== Total time ========")
    for k, v in total_time.items():
        print(f"{k}: {v:.2f} seconds")
    total = sum(total_time.values())
    if total > 60:
        print(f"\nTotal: {total:.2f} seconds ({total/60:.2f} minutes)")
    else:
        print(f"\nTotal: {total:.2f} seconds")
    print(f"============================\n")


class PmappingsOneEinsum:
    def __init__(self, einsum_name: str, pm_group_list: list[PmappingGroup]):
        self.einsum_name: str = einsum_name
        self.pmapping_groups: list[PmappingGroup] = pm_group_list
        self.tensor_names: set[str] = set(pm_group_list[0].tensor_names)

    def __getitem__(self, i):
        return self.pmapping_groups[i]


def make_full_equivalent_rank_variables(pairwise_equivalent_rank_variables):
    full_equivalent_rank_variables = {
        k: set(v) for k, v in pairwise_equivalent_rank_variables.items()
    }
    changed = True
    while changed:
        changed = False
        for r in full_equivalent_rank_variables:
            for r2 in list(full_equivalent_rank_variables[r]):
                for r3 in list(full_equivalent_rank_variables[r2]):
                    if r3 in full_equivalent_rank_variables[r]:
                        continue
                    changed = True
                    full_equivalent_rank_variables[r].add(r3)
    return full_equivalent_rank_variables


def quick_join(
    pmapping_groups: dict[str, PmappingGroup],
    mapspace_globals: MapspaceGlobals,
):
    resource2capacity = mapspace_globals.resource2capacity
    pairwise_equivalent_rank_variables = mapspace_globals.pairwise_equivalent_ranks
    aliased_tensors = mapspace_globals.aliased_tensors
    full_equivalent_rank_variables = mapspace_globals.full_equivalent_ranks

    n_mappings = {}
    runtime = {}
    nbuckets = []

    n_evaluations = 0

    pmapping_groups = list(pmapping_groups.items())

    init_print_time()

    pmapping_groups = [PmappingsOneEinsum(*s) for s in pmapping_groups]

    if not pmapping_groups:
        raise ValueError("No PmappingGroups to join")

    # ======================================================================
    # Initial consolidate and group all PmappingGroups
    # ======================================================================
    for i, sim_holder in enumerate(pmapping_groups):
        right_tensors = set.union(
            set(), *[s.tensor_names for s in pmapping_groups[i + 1 :]]
        )
        if i == 0:
            sim_holder.pmapping_groups = PmappingGroup.left_consolidate(
                sim_holder.pmapping_groups,
                right_tensors,
            )
            continue
        t0 = time.time()
        left_tensors = set.union(set(), *[s.tensor_names for s in pmapping_groups[:i]])
        live_tensors = right_tensors
        shared_tensors = left_tensors & sim_holder.tensor_names
        sim_holder.pmapping_groups = sorted(
            sim_holder.pmapping_groups, key=lambda x: len(x.mappings.data), reverse=True
        )
        sim_holder.pmapping_groups = PmappingGroup.right_consolidate(
            sim_holder.pmapping_groups,
            live_tensors,
            shared_tensors,
        )
        sim_holder.pmapping_groups = PmappingGroup.combine_combineable(
            sim_holder.pmapping_groups,
            left_tensors | right_tensors,
        )
        if i > 0:
            sim_holder.pmapping_groups = PmappingGroup.group_right(
                sim_holder.pmapping_groups, left_tensors, drop_tags=True
            )
        einsum, prev_einsum = sim_holder.einsum_name, pmapping_groups[i - 1].einsum_name
        runtime[f"{prev_einsum} → {einsum}"] = time.time() - t0
        t0 = time.time()

    n_iterations = 0
    total_iterations = len(pmapping_groups)

    def grab_sim_holder() -> (
        tuple[dict[Compatibility, list[PmappingGroup]], str, set[str]]
    ):
        nonlocal n_iterations
        n_iterations += 1
        holder = pmapping_groups.pop(0)
        return holder.pmapping_groups, holder.einsum_name, holder.tensor_names

    if pmapping_groups:
        left, left_einsum, left_tensors = grab_sim_holder()

    partial_mapping_size = 1
    while pmapping_groups:
        t0 = time.time()
        # ======================================================================
        # Grab new Einsum from the right. Record logging data and find still
        # tensors that will be live after this Einsum.
        # ======================================================================
        nbuckets.append(len(left))
        # nmappings.append(sum(len(s.mappings.data) for s in left))
        right, right_einsum, right_tensors = grab_sim_holder()
        right_rank_variables = mapspace_globals.einsum2ranks[right_einsum]

        partial_mapping_size += 1

        live_tensors = set.union(set(), *[s.tensor_names for s in pmapping_groups])
        shared_tensors = set(left_tensors) & set(right_tensors)
        live_tensors_with_right = live_tensors | right_tensors

        # ======================================================================
        # Clean up the previously-combined PmappingGroups. Consolidate, combine, group
        # them into buckets.
        # ======================================================================

        left = PmappingGroup.combine_combineable(
            left,
            live_tensors | right_tensors,
        )

        # Group left and right into buckets
        left = PmappingGroup.group_left(left, right_tensors, drop_tags=True)

        # ======================================================================
        # Remove dead tensors from left and right. This happens after grouping
        # because we only reserve space for shared tensors after it's dead. This
        # is in case the tensor lifetime extends beyond the Einsums for which it
        # is used.
        # ======================================================================
        PmappingGroup.remove_dead_tensors(
            [s for lr in [left, right] for v in lr.values() for s in v], live_tensors
        )

        # ======================================================================
        # Merge the left and right buckets.
        # ======================================================================
        combined: list[PmappingGroup] = []
        for k in left:
            for k_translated in get_possible_translations(
                k,
                pairwise_equivalent_rank_variables,
                full_equivalent_rank_variables,
                right_rank_variables,
            ):
                for a, b in itertools.product(left[k], right.get(k_translated, [])):
                    if a.compatibility.tags.are_compatible_with(b.compatibility.tags):
                        combined.append(
                            a.merge_next(
                                b,
                                live_tensors,
                                live_tensors_with_right,
                                aliased_tensors,
                                resource2capacity,
                                delay=False,
                            )
                        )

        if not combined:
            raise ValueError("No match found for any group")

        # ======================================================================
        # Update left for the next iteration.
        # =================================================================
        left = combined
        left_einsum = right_einsum
        left_tensors |= right_tensors

    # ======================================================================
    # Final consolidate and group
    # ======================================================================
    t0 = time.time()
    left = PmappingGroup.left_consolidate(left, None)
    s_final = PmappingGroup.combine_combineable(left, set(), drop_tags=True)
    assert len(s_final) == 1
    mappings = s_final[0].mappings

    return mappings



import copy
from dataclasses import dataclass, field
from accelforge.frontend.arch import Network as NetworkSpec
from accelforge.frontend.mapping import (
    Compute,
    Mapping,
    Toll,
    Reservation,
    Spatial,
    Storage,
    Temporal,
)
from typing import Any, List

from accelforge.frontend import arch
import accelforge.frontend.mapping as mapping_spec
from accelforge.frontend.mapping import (
    Mapping,
    MappingNode,
    Spatial,
    Temporal,
    Storage,
    Reservation,
    Loop,
    TensorHolder,
    Toll,
)
from accelforge.frontend.workload import (
    Workload,
    TensorName,
)
from accelforge.frontend._workload_isl._symbolic import (
    get_projection_expr,
    get_rank_variable_relevancy,
    compute_dense_tile_occupancy,
    Irrelevant,
    Relevant,
    PartiallyRelevant,
)

from accelforge.model._looptree.types import Buffet, Compute, Network
from accelforge.model._looptree.reuse.symbolic.mapping_utils import (
    DataMovementConnections,
    get_tensor_to_backer_id,
)

from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job
from accelforge.util._sympy.broadcast_max import Min, Max, MaxGeqZero, MinGeqZero
from accelforge.mapper.FFM._pareto_df.df_convention import iterations2col

import sympy

SYMBOL = "symbol"
IMPERFECT = False

PRINT_FORMULAS = False


class Uninitialized:
    def __init__(self):
        pass

    def __str__(self):
        return "Uninitialized"

    def __repr__(self):
        return "Uninitialized()"

    def __rmul__(self, other):
        return self * other

    def __mul__(self, other):
        return self

    def __radd__(self, other):
        return self + other

    def __add__(self, other):
        return self


# TODO: unsure if this is needed. If the sympy symbol is created with the
# correct assumption (e.g., positive), this should be automatic.
def min_nonzero(a: Any, b: Any) -> Any:
    if a == 0:
        return b
    if b == 0:
        return a
    return MinGeqZero(a, b)


def max_dict(a: dict[Any, Any], b: dict[Any, Any]) -> dict[Any, Any]:
    new = {**a}
    for key, value in b.items():
        new[key] = MaxGeqZero(new[key], value) if key in new else value
    assert isinstance(new, dict)
    return new


@dataclass
class NetworkStats:
    total_hops: Any = field(default=0)
    max_hops: Any = field(default=0)

    def repeat(self, n_repeats):
        new = copy.copy(self)
        new.total_hops *= n_repeats
        return new

    def combine(self, other: "NetworkStats"):
        self.total_hops += other.total_hops
        self.max_hops = max(self.max_hops, other.max_hops)


@dataclass
class BuffetStats:
    total_reads_to_parent: Any = field(default=0)
    total_writes_to_parent: Any = field(default=0)
    max_per_parent_reads_to_parent: Any = field(default=0)
    max_per_parent_writes_to_parent: Any = field(default=0)

    total_reads_to_peer: Any = field(default=0)
    total_writes_to_peer: Any = field(default=0)
    max_per_unit_reads_to_peer: Any = field(default=0)
    max_per_unit_writes_to_peer: Any = field(default=0)

    # Skip the first iteration of temporal loops for data that is written
    total_skipped_first_reads_to_parent: Any = field(default=0)
    total_skipped_first_reads_to_peer: Any = field(default=0)
    min_per_parent_skipped_first_reads_to_parent: Any = field(default=0)
    min_per_unit_skipped_first_writes_to_peer: Any = field(default=0)

    max_occupancy: Any = field(default=0)
    _n_loops_above: int = field(default=0)

    # These are used to calculate energy and latency
    total_write_actions: Any = field(default=0)
    max_per_unit_write_actions: Any = field(default=0)
    total_read_actions: Any = field(default=0)
    max_per_unit_read_actions: Any = field(default=0)

    total_skipped_first_write_actions: Any = field(default=0)
    min_per_unit_skipped_first_write_actions: Any = field(default=0)
    total_skipped_first_read_actions: Any = field(default=0)
    min_per_unit_skipped_first_read_actions: Any = field(default=0)

    persistent: bool = field(default=False)

    @property
    def n_loops_above(self) -> int:
        if self.persistent:
            return -1
        return self._n_loops_above

    @n_loops_above.setter
    def n_loops_above(self, value: int):
        self._n_loops_above = value

    def repeat_temporal(self, factor: int, is_fully_relevant: bool) -> "BuffetStats":
        new = copy.copy(self)
        for attr in self.__dict__:
            if not attr.startswith(("total_", "max_", "min_")):
                continue
            if "skipped_first" in attr and not is_fully_relevant:
                continue  # First actions occur once per relevant iteration.
            if attr == "max_occupancy":
                continue  # Max occupancy is not affected by temporal loops above
            setattr(new, attr, getattr(new, attr) * factor)
        return new

    def repeat_spatial(self, factor: int, reuse_parent_accesses: bool) -> "BuffetStats":
        """
        Duplicate attributes to account for spatial fanout.

        Paramters
        ---------
        factor:
            duplication factor
        reuse_parent_access:
            whether to reuse accesses to parent (if True, multicast is assumed
            and accesses to parents are not duplicated).
        """
        new = copy.copy(self)
        for attr in self.__dict__:
            if not attr.startswith(("total_", "max_", "min_")):
                continue
            if "parent" in attr and reuse_parent_accesses:
                continue  # If parent accesses are reused, no need to multiply
            if "per_unit" in attr:
                continue  # Spatial fanout doesn't affect per-unit stats
            if attr == "max_occupancy":
                continue  # Max occupancy is not affected by temporal loops above
            setattr(new, attr, getattr(new, attr) * factor)
        return new

    def max(self, **kwargs: Any):
        for key, value in kwargs.items():
            setattr(self, key, MaxGeqZero(getattr(self, key), value))

    def min(self, **kwargs: Any):
        for key, value in kwargs.items():
            setattr(self, key, MinGeqZero(getattr(self, key), value))

    def __add__(self, other: "BuffetStats") -> "BuffetStats":
        new = copy.copy(self)
        for attr in self.__dict__:
            if attr.startswith("min_"):
                setattr(
                    new, attr, min_nonzero(getattr(self, attr), getattr(other, attr))
                )
            elif attr.startswith("max_"):
                setattr(
                    new, attr, MaxGeqZero(getattr(self, attr), getattr(other, attr))
                )
            elif attr.startswith("total_"):
                setattr(new, attr, getattr(self, attr) + getattr(other, attr))
            elif getattr(self, attr) is None:
                setattr(new, attr, getattr(other, attr))
            elif getattr(other, attr) is None:
                setattr(new, attr, getattr(self, attr))
            else:
                assert getattr(self, attr) == getattr(
                    other, attr
                ), f"BUG: {attr} is different. self: {getattr(self, attr)} other: {getattr(other, attr)}"
        return new

    def __iadd__(self, other: "BuffetStats") -> "BuffetStats":
        new = self + other
        for key, value in new.__dict__.items():
            setattr(self, key, value)
        return self

    def net_total_read_actions(self) -> Any:
        return self.total_read_actions - self.total_skipped_first_read_actions

    def net_total_write_actions(self) -> Any:
        return self.total_write_actions - self.total_skipped_first_write_actions

    def net_max_per_unit_read_actions(self) -> Any:
        return (
            self.max_per_unit_read_actions
            - self.min_per_unit_skipped_first_read_actions
        )

    def net_max_per_unit_write_actions(self) -> Any:
        return (
            self.max_per_unit_write_actions
            - self.min_per_unit_skipped_first_write_actions
        )

    @classmethod
    def blank(cls):
        stats = cls()
        stats.n_loops_above = None  # Inherit from whoever is added to this
        return stats


@dataclass
class ComputeStats:
    total_ops: Any = field(default=0)
    max_per_unit_ops: Any = field(default=0)
    # "max" below refers to the longest latency of any iteration
    max_latency: Any = field(default=0)
    # Mapping from the loop-index (0 at top) to the latency of the first
    # iteration of that loop. "Max" because we may have loops above that and we
    # will take the maximum of the firsts.
    max_first_latency: dict[int, Any] = field(default_factory=dict)

    def repeat_temporal(self, factor: int) -> "ComputeStats":
        new = copy.copy(self)
        new.total_ops *= factor
        new.max_per_unit_ops *= factor
        new.max_latency *= factor
        # NOTE: max_first_latency does not change
        return new

    def repeat_spatial(self, factor: int) -> "ComputeStats":
        new = copy.copy(self)
        new.total_ops *= factor
        return new

    def __add__(self, other: "ComputeStats") -> "ComputeStats":
        new = copy.copy(self)
        new.total_ops += other.total_ops
        new.max_per_unit_ops += other.max_per_unit_ops
        new.max_latency += other.max_latency
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        new.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY
        return new

    def combine_temporal(self, other: "ComputeStats"):
        self.total_ops += other.total_ops
        self.max_per_unit_ops += other.max_per_unit_ops
        self.max_latency += other.max_latency
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        self.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY

    def combine_spatial(self, other: "ComputeStats"):
        self.total_ops += other.total_ops
        self.max_per_unit_ops = MaxGeqZero(
            self.max_per_unit_ops, other.max_per_unit_ops
        )
        self.max_latency = MaxGeqZero(self.max_latency, other.max_latency)
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        self.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY


@dataclass
class SymbolicAnalysisOutput:
    compute_stats: dict[Compute, ComputeStats] = field(default_factory=dict)

    buffet_stats: dict[Buffet, BuffetStats] = field(default_factory=dict)

    network_stats: dict[Network, NetworkStats] = field(default_factory=dict)

    # Mapping [level, einsum] to the fanout
    fanout: dict[(Buffet, str), int] = field(default_factory=dict)

    # Mapping [einsum] to the number of temporal steps
    temporal_steps: dict[str, int] = field(default_factory=dict)

    symbols: list[sympy.Symbol] = field(default_factory=list)

    # tensor to the mapping for that particular tensor
    tensor2mapping: dict[TensorName, Mapping] = field(default_factory=dict)

    def get_buffet_for_tensor(self, tensor: TensorName) -> Buffet:
        for buffet in self.buffet_stats:
            if buffet.tensor == tensor:
                return buffet
        raise ValueError(f"Buffet for tensor {tensor} not found")

    def max(self, **kwargs: Any):
        for key, value in kwargs.items():
            assert key in [
                "compute_stats",
                "stats",
                "fanout",
                "temporal_steps",
            ]
            previous = getattr(self, key)
            for k, v in value.items():
                previous.setdefault(k, {})
                for k2, v2 in v.items():
                    if k2 in previous[k]:
                        previous[k][k2] = MaxGeqZero(previous[k][k2], v2)
                    else:
                        previous[k][k2] = v2

    def get_child_buffet_stats(self, buffet: Buffet) -> BuffetStats:
        seen = False
        for child_buffet, child_stats in reversed(self.buffet_stats.items()):
            if not seen:
                seen = child_buffet == buffet
                continue
            if child_buffet.tensor == buffet.tensor:
                return child_stats
        return None

    def sum_buffet_stats_per_level(self) -> dict[str, BuffetStats]:
        result: dict[str, BuffetStats] = {}
        for buffet, stats in self.buffet_stats.items():
            result.setdefault(buffet.level, BuffetStats.blank())
            result[buffet.level] += stats
        return result

    def add_buffet_stats_and_symbols(self, other: "SymbolicAnalysisOutput"):
        assert not (set(self.buffet_stats) & set(other.buffet_stats)), "BUG"
        self.buffet_stats.update(other.buffet_stats)
        # if self.temporal_steps != other.temporal_steps:
        #     print(f'Temporal steps are different.')
        #     print(f'\tmine:  {self.temporal_steps}')
        #     print(f'\tother: {other.temporal_steps}')
        # assert self.temporal_steps == other.temporal_steps, "BUG"
        self.temporal_steps.update(other.temporal_steps)
        self.symbols.extend([s for s in other.symbols if s not in self.symbols])
        # Assert compute stats are the same
        # assert self.compute_stats == other.compute_stats, "BUG"

    def add_network_stats(self, other: "SymbolicAnalysisOutput"):
        assert not (set(self.network_stats) & set(other.network_stats)), "BUG"
        self.network_stats.update(other.network_stats)


@dataclass
class AnalysisInfo:
    """Information needed within the analysis step by multiple functions that
    can be computed once at the beginning.
    """

    mapping: Mapping
    workload: Workload
    full_rank_variable_shapes: dict
    all_tensors: set

    einsum_tensor_to_projection: dict
    tensor_to_relevancy: dict
    tensor_to_backer_id: dict[TensorName, int]

    is_copy_operation: TensorName | None

    job: Job

    tensor_to_reservation_backer_id: dict[TensorName, int] = field(default_factory=dict)

    data_movement_connections: DataMovementConnections = None

    # We track first latency for these nodes (should be Temporal)
    last_temporal_node_idx: int = None
    """
    node idx of the last (above) temporal node
    """
    idxs_to_track_first_latency: set[int] = field(default_factory=set)
    """
    node idxs for which we track first latency
    """


def quick_insert_reservation_nodes(job: Job) -> list[MappingNode]:
    mapping = list(job.mapping.nodes)
    workload = job.spec.workload

    # TODO: Subclass reservation with TensorReservation or something so that we can
    # track which are for tensors and which are for non-tensor resources.

    info = AnalysisInfo(
        mapping=None,
        workload=workload,
        full_rank_variable_shapes=None,
        all_tensors=None,
        einsum_tensor_to_projection=None,
        tensor_to_relevancy=job.tensor_to_relevancy,
        tensor_to_backer_id=None,
        is_copy_operation=None,
        job=None,
    )
    insert_reservation_nodes(mapping, info, job.fusable_tensors)
    m = Mapping(nodes=mapping)
    m._n_loop_orders = job.mapping._n_loop_orders
    return m


def convert_to_copy(
    mapping: list[MappingNode], workload: Workload
) -> tuple[list[MappingNode], dict[TensorName, int]]:
    mapping = copy.deepcopy(mapping)

    # Calculate this BEFORE we modify the mapping. We're going to have the copy source
    # tensor moving upward sometimes, and we don't want the backing tensor holder
    tensor_to_backer_id = get_tensor_to_backer_id(mapping)

    first_input_tensor = workload.einsums[mapping[-1].einsum].copy_source_tensor()

    for node in mapping:
        if isinstance(node, TensorHolder):
            if node.tensors:
                node.tensors = [first_input_tensor]
                node._lower = False

    to_remove = []
    i = 0
    while i < len(mapping):
        node = mapping[i]
        if isinstance(node, TensorHolder):
            j = i + 1
            while j < len(mapping):
                node2 = mapping[j]
                if (
                    isinstance(node2, TensorHolder)
                    and node.component == node2.component
                ):
                    mapping.pop(j)
                else:
                    j += 1
        i += 1
    mapping = [node for node in mapping if node not in to_remove]

    return mapping, tensor_to_backer_id


def analyze_reuse_and_add_reservations_to_mapping(
    job: Job,
    add_reservations: bool = True,
) -> SymbolicAnalysisOutput:
    mapping = job.mapping.nodes
    workload = job.spec.workload
    einsum_name = mapping[-1].einsum

    is_copy_operation = workload.einsums[einsum_name].is_copy_operation
    symbols = insert_sympy_symbols(job.mapping.nodes, job)

    if is_copy_operation:
        mapping, tensor_to_backer_id = convert_to_copy(mapping, workload)
    else:
        tensor_to_backer_id = get_tensor_to_backer_id(mapping)

    if add_reservations:
        job.mapping = quick_insert_reservation_nodes(job)
    # print(f'Job mapping: {job.mapping.compact_str()}')
    # for n in job.mapping.nodes:
    #     print(f'\t{n.compact_str()}')

    einsum_tensor_to_projection = {}
    einsum = workload.einsums[einsum_name]
    all_tensors = einsum.tensor_names
    for tensor in all_tensors:
        einsum_tensor_to_projection[(einsum_name, tensor)] = get_projection_expr(
            einsum, tensor
        )
    tensor_to_relevancy = {
        tensor: get_rank_variable_relevancy(einsum, tensor) for tensor in all_tensors
    }
    assert all_tensors, f"Einsum {einsum_name} has no tensors"

    """
    Note for how this works.

    Spatial loops are weird, because they don't belong at a single point in the loop
    nest. For example:

    - DRAM keep A, B
    - *
    - Reg keep A
    - for n in [0..N)
    - GLB keep B
    - *
    - Compute

    A loop 'spatial-for (Reg) k in [0..K)' would affect the register at the point of the
    first asterisk, but the global buffer at the point of the second asterisk.

    To handle this, we make a separate mapping for each tensor, analyze each, and
    combine the results.

    To anyone who would like to create behavior that simultaneously looks at multiple
    storage nodes for a given memory, note that there will be two challenges to address:

    1. The code currently analyzes one tensor at a time. This could be fixed by
       processing all mapping(s) together, applying loop(s) from each to only the
       appropriate nodes.
    2. The code must analyze one storage node at a time, and there may be temporal and
       spatial nodes between two storage nodes for a given memory, which would separate
       the analysis steps for the storage nodes. This may be addressed by only
       performing such analysis until the outermost storage node for a particular memory
       has been analyzed.
    """
    result = None

    tensor2mapping = {}
    index_expressions = set(einsum.indexing_expressions)
    for k, v in job.rank_variable_bounds.items():
        index_expressions.add(f"0 < {k} <= {v}")
    for tensor in all_tensors:
        cur_mapping = job.mapping._get_single_tensor_mapping(
            tensor, job.flattened_arch, index_expressions
        )
        info = AnalysisInfo(
            mapping=cur_mapping.nodes,
            workload=workload,
            full_rank_variable_shapes=job.rank_variable_bounds,
            all_tensors=set([tensor]),
            einsum_tensor_to_projection=einsum_tensor_to_projection,
            tensor_to_relevancy=tensor_to_relevancy,
            tensor_to_backer_id=tensor_to_backer_id,
            is_copy_operation=is_copy_operation,
            job=job,
            data_movement_connections=DataMovementConnections.from_pmapping(
                cur_mapping.nodes
            ),
        )
        cur_result = analyze_node(0, job.rank_variable_bounds, info)
        if result is None:
            result = cur_result
        else:
            result.add_buffet_stats_and_symbols(cur_result)
            result.add_network_stats(cur_result)
        tensor2mapping[tensor] = cur_mapping

    result.symbols = symbols
    result.tensor2mapping = tensor2mapping

    if PRINT_FORMULAS:
        print(f"Mapping:")
        for node in mapping:
            print(f"\t{node.compact_str()}")

        for buffet, stats in result.buffet_stats.items():
            print(f"Einsum {buffet.einsum} tensor {buffet.tensor} level {buffet.level}")
            print(f"\tTotal reads to parent: {stats.total_reads_to_parent}")
            print(f"\tTotal writes to parent: {stats.total_writes_to_parent}")
            print(
                f"\tMax per parent reads to parent: {stats.max_per_parent_reads_to_parent}"
            )
            print(
                f"\tMax per parent writes to parent: {stats.max_per_parent_writes_to_parent}"
            )
            print(f"\tTotal reads to peer: {stats.total_reads_to_peer}")
            print(f"\tTotal writes to peer: {stats.total_writes_to_peer}")
            print(f"\tMax per unit reads to peer: {stats.max_per_unit_reads_to_peer}")
            print(f"\tMax per unit writes to peer: {stats.max_per_unit_writes_to_peer}")

    return result


class ReservationAnalysisTracker:
    def __init__(self, buffet, node):
        self.buffet: Buffet = buffet
        self.node: TensorHolder = node

        # These are interface (TODO: should be property)
        self.is_fill_level = False
        self.should_stop = False
        self.insert_reservation_under = False
        self.insert_fill_under = False

        # Temporary values
        self.has_filled = False

    def track_temporal_loop(self, relevancy, node):
        self.is_fill_level = False
        self.insert_reservation_under = False
        self.insert_fill_under = False

        if isinstance(relevancy, Irrelevant):
            if not self.has_filled:
                self.is_fill_level = True
                self.has_filled = True

            self.should_stop = True
        elif isinstance(relevancy, Relevant):
            self.should_stop = False
        elif isinstance(relevancy, PartiallyRelevant):
            self.last = True

            if not self.has_filled:
                self.is_fill_level = True
                self.has_filled = True

            self.should_stop = True
            self.insert_reservation_under = True
        else:
            raise ValueError(f"Unknown relevancy {relevancy}")

    def track_compute(self):
        self.should_stop = True
        if not self.has_filled:
            self.is_fill_level = True
            self.has_filled = True

    def track_spatial_loop(self, relevancy, node):
        if node.component != self.buffet.level:
            self.should_stop = True
            if not self.has_filled:
                self.is_fill_level = True
                self.has_filled = True
            return

        self.is_fill_level = False
        self.should_stop = False


def insert_reservation_nodes(
    mapping, info: AnalysisInfo, fusable_tensors: set[TensorName]
):
    trackers: list[ReservationAnalysisTracker] = []
    einsum = info.workload.einsums[mapping[-1].einsum]
    non_intermediate_tensors = (
        einsum.tensor_names - info.workload.tensor_names_used_in_multiple_einsums
    )
    seen_tensors = set()  # reservation for top-level buffets cannot be lowered

    n_nodes = len(mapping)
    i = 0
    while i < n_nodes:
        node = mapping[i]
        to_remove = []
        if isinstance(node, Reservation):
            pass
        elif isinstance(node, Temporal):
            rank = node.rank_variable
            for tracker in trackers:
                relevancy = info.tensor_to_relevancy[tracker.buffet.tensor]
                tracker.track_temporal_loop(relevancy[rank], node)
        elif isinstance(node, Spatial):
            rank = node.rank_variable
            for tracker in trackers:
                relevancy = info.tensor_to_relevancy[tracker.buffet.tensor]
                tracker.track_spatial_loop(relevancy[rank], node)
        elif isinstance(node, TensorHolder):
            for tracker in trackers:
                tracker.should_stop = True
                tracker.insert_reservation_under = False
            for tensor in node.tensors:
                tensor = TensorName(tensor)
                buffet = Buffet(tensor, mapping[-1].einsum, node.component)
                trackers.append(ReservationAnalysisTracker(buffet, node))
                if not node._lower or tensor not in seen_tensors:
                    seen_tensors.add(tensor)
                    trackers[-1].is_fill_level = True
                    trackers[-1].insert_reservation_under = True
                    trackers[-1].insert_fill_under = True
                    trackers[-1].should_stop = True
        elif isinstance(node, mapping_spec.Compute):
            for tracker in trackers:
                tracker.track_compute()
                tracker.insert_reservation_under = False
        else:
            raise NotImplementedError(f"Unknown node type {type(node)}")

        reservation_insert_below = []
        reservation_insert_above = []
        for j in range(len(trackers) - 1, -1, -1):
            if not trackers[j].should_stop:
                continue
            tracker = trackers.pop(j)
            buffet = tracker.buffet
            node = Reservation(purposes=[buffet.tensor], resource=buffet.level)
            node.persistent = tracker.node.persistent
            node._backing = tracker.node._backing

            if (
                buffet.tensor not in info.tensor_to_reservation_backer_id
                and buffet.tensor in info.workload.tensor_names_used_in_multiple_einsums
            ):
                info.tensor_to_reservation_backer_id[buffet.tensor] = id(node)

            if tracker.insert_reservation_under:
                reservation_insert_below.append(node)
            else:
                reservation_insert_above.append(node)

        # The order of these for loops is important. Reservation must be below fill.
        for node in reservation_insert_below:
            mapping.insert(i + 1, node)
            i += 1
        for node in reservation_insert_above:
            mapping.insert(i, node)
            i += 1

        i += 1
        n_nodes = len(mapping)

    label_fused_loops(mapping, fusable_tensors)


def label_fused_loops(mapping: List[MappingNode], fusable_tensors: set[TensorName]):
    last_backer = -1
    assert isinstance(
        fusable_tensors, set
    ), f"Fusable tensors must be a set, got {type(fusable_tensors)}"

    for i, node in enumerate(mapping):
        if not isinstance(node, TensorHolder) or isinstance(node, Reservation):
            continue
        assert isinstance(node._backing, set)
        if node._backing & fusable_tensors:
            last_backer = i
            # We may lower this node through at most one additional loop
            if node._lower:
                last_backer += 1
    if last_backer == -1 and fusable_tensors:
        raise ValueError(
            f"No backing TensorHolder found in mapping {", ".join(m.compact_str() for m in mapping)}"
        )

    for i, node in enumerate[MappingNode](mapping):
        if isinstance(node, Loop):
            node._fused = i < last_backer

    return mapping


def analyze_node(node_idx, current_shape, info: AnalysisInfo) -> SymbolicAnalysisOutput:
    node = info.mapping[node_idx]
    class2analysis_function = {
        Temporal: analyze_temporal,
        Spatial: analyze_spatial,
        Storage: analyze_storage,
        Reservation: analyze_reservation,
        mapping_spec.Compute: analyze_compute,
        Toll: analyze_toll,
    }
    if type(node) not in class2analysis_function:
        raise TypeError(f"Unknown node type {type(node)}")
    return class2analysis_function[type(node)](node_idx, current_shape, info)


def analyze_temporal(
    node_idx, current_shape, info: AnalysisInfo
) -> SymbolicAnalysisOutput:
    mapping = info.mapping
    node = mapping[node_idx]
    stride_and_shape = get_stride_and_tile_shape(node, current_shape, node_idx, info)

    result_accumulator = SymbolicAnalysisOutput()

    first_latency = None

    def handle_repeated_value(repeated_shape):
        nonlocal first_latency
        shape_value = repeated_shape.value
        shape_repeats = repeated_shape.repeats

        child_shape = current_shape.copy()
        child_shape[node.rank_variable] = shape_value

        child_result = analyze_node(node_idx + 1, child_shape, info)

        accumulated_buffet_stats = result_accumulator.buffet_stats
        for buffet, stats in child_result.buffet_stats.items():
            relevancy = info.tensor_to_relevancy[buffet.tensor][node.rank_variable]
            is_fully_relevant = isinstance(relevancy, Relevant)
            accumulated_stats = accumulated_buffet_stats.setdefault(
                buffet, BuffetStats.blank()
            )
            accumulated_stats += stats.repeat_temporal(
                shape_repeats, is_fully_relevant=is_fully_relevant
            )
            accumulated_stats.n_loops_above = stats.n_loops_above + 1

        for network, network_stats in child_result.network_stats.items():
            result_accumulator.network_stats[network] = network_stats.repeat(
                shape_repeats
            )

        for einsum, child_steps in child_result.temporal_steps.items():
            if einsum not in result_accumulator.temporal_steps:
                result_accumulator.temporal_steps[einsum] = 0
            result_accumulator.temporal_steps[einsum] += child_steps * shape_repeats

        result_accumulator.max(fanout=child_result.fanout)

        for key in child_result.compute_stats:
            if first_latency is None:
                first_latency = child_result.compute_stats[key].max_latency

            compute_stats = result_accumulator.compute_stats.setdefault(
                key, ComputeStats()
            )
            compute_stats += child_result.compute_stats[key].repeat_temporal(
                shape_repeats
            )
            result_accumulator.compute_stats[key] = compute_stats

    info.last_temporal_node_idx = node_idx

    shape = stride_and_shape.shape
    if isinstance(shape, SequenceOfRepatedvalues):
        for repeated_shape in shape.sequence:
            assert isinstance(repeated_shape, RepeatedValue)
            handle_repeated_value(repeated_shape)
    elif isinstance(shape, RepeatedValue):
        handle_repeated_value(shape)

    if node_idx in info.idxs_to_track_first_latency:
        for compute_stat in result_accumulator.compute_stats.values():
            # Should be the first time we store this value
            assert node_idx not in compute_stat.max_first_latency
            compute_stat.max_first_latency[node_idx] = first_latency

    return result_accumulator


def analyze_spatial(node_idx, current_shape, info: AnalysisInfo):
    mapping = info.mapping
    einsum_name = mapping[-1].einsum
    node: Spatial = mapping[node_idx]
    rank_var = node.rank_variable
    node_dim = node.name
    spatial_component = find_component_object(node.component, info.job.flattened_arch)
    component_spatial_dim = spatial_component.spatial[node_dim]
    stride_and_shape = get_stride_and_tile_shape(node, current_shape, node_idx, info)

    result_accumulator = SymbolicAnalysisOutput()

    overall_max_hops = 0

    def handle_repeated_value(repeated_shape):
        nonlocal overall_max_hops
        shape_value = repeated_shape.value
        shape_repeats = repeated_shape.repeats

        child_shape = current_shape.copy()
        child_shape[node.rank_variable] = shape_value

        child_result = analyze_node(node_idx + 1, child_shape, info)

        accumulated_buffet_stats = result_accumulator.buffet_stats
        child_stats = list(child_result.buffet_stats.items())
        for i, (buffet, buffet_stats) in enumerate(child_stats):
            stats = buffet_stats
            accumulated_stats = accumulated_buffet_stats.setdefault(
                buffet, BuffetStats.blank()
            )
            relevancy = info.tensor_to_relevancy[buffet.tensor][rank_var]

            # Reuse parent accesses only:
            # - Irrelevant loops
            # - The outermost level that holds the tensor (the one whose parent accesses
            #   will be going through the network)
            last_buffet = True
            for other_buffet, _ in child_stats[i + 1 :]:
                if other_buffet.tensor == buffet.tensor:
                    last_buffet = False
                    break

            reuse_parent_accesses = (
                last_buffet
                and isinstance(relevancy, Irrelevant)
                and buffet.tensor in component_spatial_dim.may_reuse
            )

            accumulated_stats += stats.repeat_spatial(
                shape_repeats, reuse_parent_accesses=reuse_parent_accesses
            )
            accumulated_stats.n_loops_above = stats.n_loops_above + 1

        for network, child_network_stats in child_result.network_stats.items():
            if network not in result_accumulator.network_stats:
                result_accumulator.network_stats[network] = NetworkStats()
            accumulated_network_stats = result_accumulator.network_stats[network]

            accumulated_network_stats.total_hops += (
                child_network_stats.total_hops * shape_repeats
            )
            accumulated_network_stats.max_hops = max(
                accumulated_network_stats.max_hops,
                child_network_stats.max_hops,
            )
            projection = info.einsum_tensor_to_projection[(einsum_name, network.tensor)]
            component_object = find_component_object(
                network.component, info.job.flattened_arch
            )
            bits_per_value_scale = component_object.bits_per_value_scale[network.tensor]
            bits_per_value = (
                bits_per_value_scale
                * info.job.einsum.tensor_accesses[network.tensor].bits_per_value
            )
            bits_per_action = component_object.bits_per_action
            if bits_per_action is not None:
                actions_per_value = bits_per_value / bits_per_action
            else:
                actions_per_value = bits_per_value
            volume = (
                compute_dense_tile_occupancy(projection, child_shape)
                * actions_per_value
            )

            if info.job.spec.arch.is_above(node.component, network.component):
                continue

            last_fanout = child_result.fanout.get((node.component, einsum_name), {})
            last_fanout = last_fanout.get(node.name, 1)
            if isinstance(relevancy, Irrelevant):
                # Cost of multicasting is the cost of delivering along the dimension
                multicast_hops = (shape_repeats - 1) * last_fanout
                multicast_cost = multicast_hops * volume
                overall_max_hops += multicast_hops

                accumulated_network_stats.total_hops += multicast_cost
                accumulated_network_stats.max_hops = max(
                    accumulated_network_stats.max_hops,
                    overall_max_hops + child_network_stats.max_hops,
                )
            elif isinstance(relevancy, Relevant):
                # Cost of unicast is the cost of delivering to each point in
                # the dimension with shape as stride
                # TODO: we should use the actual stride
                total_unicast_cost = (
                    0.5 * (shape_repeats - 1) * shape_repeats * last_fanout * volume
                )
                max_unicast_hops = (shape_repeats - 1) * last_fanout
                overall_max_hops += max_unicast_hops

                accumulated_network_stats.total_hops += total_unicast_cost
                accumulated_network_stats.max_hops = max(
                    accumulated_network_stats.max_hops,
                    overall_max_hops + child_network_stats.max_hops,
                )
            elif isinstance(relevancy, PartiallyRelevant):
                raise NotImplementedError()
            else:
                raise RuntimeError(f"unhandled relevancy type {relevancy}")

        for einsum, child_steps in child_result.temporal_steps.items():
            if einsum not in result_accumulator.temporal_steps:
                result_accumulator.temporal_steps[einsum] = child_steps
            else:
                result_accumulator.temporal_steps[einsum] = MaxGeqZero(
                    result_accumulator.temporal_steps[einsum], child_steps
                )

        my_key = (node.component, einsum_name)
        child_result.fanout.setdefault(my_key, {})

        # Propagate up everything except the current level and dimension
        child_fanout = copy.deepcopy(child_result.fanout)
        target_fanout = child_fanout[my_key].pop(node_dim, 1)
        result_accumulator.max(fanout=child_fanout)

        # Prpoagate current level and dimension * shape_repeats
        child_fanout = child_result.fanout[my_key]
        fanout = result_accumulator.fanout.setdefault(my_key, {})
        fanout.setdefault(node_dim, 0)  # TODO: Assume sympy can just take in 0
        # TODO: If node_dim was missing, the original code would have omitted
        # shape_repeats. Is this correct?
        fanout[node_dim] += target_fanout * shape_repeats

        for key in child_result.compute_stats:
            # TODO: ensure that `ComputeStats()`, which is initialized ONCE, is okay to use here
            compute_stats = result_accumulator.compute_stats.setdefault(
                key, ComputeStats()
            )
            # TODO: If check omitted. This was in the original code, check history if needed.
            compute_stats.combine_spatial(
                child_result.compute_stats[key].repeat_spatial(shape_repeats)
            )

    shape = stride_and_shape.shape
    if isinstance(shape, SequenceOfRepatedvalues):
        for repeated_shape in shape.sequence:
            assert isinstance(repeated_shape, RepeatedValue)
            handle_repeated_value(repeated_shape)
    elif isinstance(shape, RepeatedValue):
        handle_repeated_value(shape)

    return result_accumulator


def reduce_dicts(dict1: dict, dict2: dict, reduce_op):
    for key in dict1:
        if key not in dict2:
            dict2[key] = dict1[key]
        else:
            dict2[key] = reduce_op(dict1[key], dict2[key])


def get_total_to_per_unit(total, max_per_unit):
    if total == 0 and max_per_unit != 0:
        raise ValueError(f"total is 0 but max_per_unit is {max_per_unit}")
    if total == 0:
        return 1
    return max_per_unit / total


def has_parent_tensor_holder(
    tensor: TensorName, node_idx: int, info: AnalysisInfo
) -> bool:
    for node in info.mapping[:node_idx]:
        if isinstance(node, TensorHolder) and tensor in node.tensors:
            return True
    return False


def find_component_object(
    component: str, flattened_arch: list[arch.Leaf]
) -> arch.TensorHolder:
    for node in flattened_arch:
        if node.name == component:
            return node
    raise ValueError(f"Component {component} not found in flattened arch")


def analyze_storage(
    node_idx: int,
    current_shape: dict[str, int],
    info: AnalysisInfo,
    propagate_child_results: bool = False,
    count_upward_movement: bool = True,
    count_downward_movement: bool = True,
    count_writes: bool = True,
):
    mapping = info.mapping
    einsum_name = mapping[-1].einsum
    node: TensorHolder = mapping[node_idx]

    child_result = analyze_node(node_idx + 1, current_shape, info)

    for tensor in node.tensors:
        tensor = TensorName(tensor)
        buffet = Buffet(tensor, einsum_name, node.component)

        # Reservations make these, and they go below the storage node, so the buffet
        # stats are already made at this point
        stats = child_result.buffet_stats[buffet]
        backer_id = info.tensor_to_backer_id[tensor]
        is_backing = backer_id == id(node)
        if node.persistent:
            stats.persistent = True
        below_backing = backer_id in [id(m) for m in mapping[:node_idx]]

        projection = info.einsum_tensor_to_projection[(einsum_name, tensor)]

        fills = compute_dense_tile_occupancy(projection, current_shape)

        child = child_result.get_child_buffet_stats(buffet)
        inherit_from_child = propagate_child_results and child is not None

        # ==============================================================================
        # Calculate the total fills and reads to parent. These propagate upward.
        # ==============================================================================

        def inherit_add(attr: str, default_value: Any = fills) -> Any:
            val = getattr(child, attr) if inherit_from_child else default_value
            setattr(stats, attr, val + getattr(stats, attr))

        if has_parent_tensor_holder(tensor, node_idx, info):
            # Initial fetch: If we're below the backing storage, fetch data from above
            # at the beginning.
            if not is_backing and below_backing:
                inherit_add("total_reads_to_parent", fills)
                inherit_add("max_per_parent_reads_to_parent", fills)

            # Data writeback. Do not writeback if it's a copy operation and we're below
            # the backing storage; data only flows upward.

            # Writeback occurs in two cases:
            # - We're at or above the backing storage, so we need to propagate our
            #   results upward to any storage nodes that will need this data.
            # - This is a written tensor, so we need to write back the written data.
            if (
                tensor in info.workload.einsums[einsum_name].output_tensor_names
                or not below_backing
            ):
                inherit_add("total_writes_to_parent")
                inherit_add("max_per_parent_writes_to_parent")

            # For read+write tensors, we skip the first fill because the data will be
            # initialized with a zero value.
            if tensor in info.workload.einsums[einsum_name].output_tensor_names:
                inherit_add("total_skipped_first_reads_to_parent")
                inherit_add("min_per_parent_skipped_first_reads_to_parent")

        # ==============================================================================
        # Convert to actions. These are not used used upward; they are used to get
        # energy and latency.
        # ==============================================================================
        component_object = find_component_object(
            node.component, info.job.flattened_arch
        )
        bits_per_value_scale = component_object.bits_per_value_scale[tensor]
        bits_per_value = (
            bits_per_value_scale
            * info.job.einsum.tensor_accesses[tensor].bits_per_value
        )
        read_bits_per_action = component_object.actions["read"].bits_per_action
        read_scale = bits_per_value / read_bits_per_action
        if count_writes:
            write_bits_per_action = component_object.actions["write"].bits_per_action
            write_scale = bits_per_value / write_bits_per_action
        else:
            write_scale = 0

        # ==========================
        # Data exchanges with parent
        if count_downward_movement:  # Parent -> Me
            stats.total_write_actions += stats.total_reads_to_parent * write_scale
            stats.max_per_unit_write_actions += (
                stats.total_reads_to_parent * write_scale
            )
            stats.total_skipped_first_write_actions += (
                stats.total_skipped_first_reads_to_parent * write_scale
            )
            stats.min_per_unit_skipped_first_write_actions += (
                stats.min_per_parent_skipped_first_reads_to_parent * write_scale
            )

        if count_upward_movement:  # Me -> Parent
            # Comment this to have the final writeback to a buffer hit both that buffer and
            # go directly to the parent without incurring another read from the buffer.
            stats.total_read_actions += stats.total_writes_to_parent * read_scale
            stats.max_per_unit_read_actions += stats.total_writes_to_parent * read_scale

        # ========================
        # Data exchanges with peer
        stats.total_read_actions += stats.total_reads_to_peer * read_scale
        stats.total_write_actions += stats.total_reads_to_peer * write_scale

        # =========================
        # Data exchanges with child
        if child is not None:
            if count_downward_movement:  # Me -> Child
                stats.total_read_actions += child.total_reads_to_parent * read_scale
                stats.max_per_unit_read_actions += (
                    child.max_per_parent_reads_to_parent * read_scale
                )
                # Skip first read
                stats.total_skipped_first_read_actions += (
                    child.total_skipped_first_reads_to_parent * read_scale
                )
                stats.min_per_unit_skipped_first_read_actions += (
                    child.min_per_parent_skipped_first_reads_to_parent * read_scale
                )

            if count_upward_movement:  # Child -> Me
                stats.total_write_actions += child.total_writes_to_parent * write_scale
                stats.max_per_unit_write_actions += (
                    child.max_per_parent_writes_to_parent * write_scale
                )

    return child_result


def analyze_toll(node_idx, current_shape, info: AnalysisInfo):
    mapping = info.mapping
    einsum_name = mapping[-1].einsum
    node = mapping[node_idx]
    component_object = find_component_object(node.component, info.job.flattened_arch)
    storage_result = analyze_storage(
        node_idx,
        current_shape,
        info,
        propagate_child_results=True,
        count_upward_movement=component_object.direction != "down",
        count_downward_movement=component_object.direction != "up",
        count_writes=False,
    )
    for tensor in node.tensors:
        buffet = Buffet(tensor, einsum_name, node.component)
        stats = storage_result.buffet_stats[buffet]
        stats.max_occupancy = 0
        assert stats.total_write_actions == 0
    return storage_result


def analyze_reservation(node_idx, current_shape, info: AnalysisInfo):
    mapping = info.mapping
    einsum_name = mapping[-1].einsum
    node = mapping[node_idx]
    tensor = TensorName(node.purpose)

    if info.last_temporal_node_idx is not None and id(
        node
    ) == info.tensor_to_reservation_backer_id.get(node.purpose, None):
        info.idxs_to_track_first_latency.add(info.last_temporal_node_idx)

    child_result = analyze_node(node_idx + 1, current_shape, info)

    buffet = Buffet(tensor, einsum_name, node.resource)

    # Reservation nodes are the first to produce stats for a buffet
    assert buffet not in child_result.buffet_stats

    stats = BuffetStats()
    projection = info.einsum_tensor_to_projection[(einsum_name, tensor)]
    component_object = find_component_object(node.resource, info.job.flattened_arch)
    bits_per_value_scale = component_object.bits_per_value_scale[tensor]
    bits_per_value = (
        bits_per_value_scale * info.job.einsum.tensor_accesses[tensor].bits_per_value
    )
    stats.max_occupancy = (
        compute_dense_tile_occupancy(projection, current_shape) * bits_per_value
    )
    child_result.buffet_stats[buffet] = stats

    # Reservation nodes are the first to produce stats for a network
    network_node = info.job.spec.arch.find_first_of_type_above(
        NetworkSpec, buffet.level, default=None
    )
    if network_node is not None:
        network = Network(
            tensor,
            einsum_name,
            info.data_movement_connections.get_src(buffet),
            buffet,
            component=network_node.name if network_node else network_node,
        )
        assert network not in child_result.network_stats
        child_result.network_stats[network] = NetworkStats()

    fanout_key = (node.resource, einsum_name)
    if fanout_key not in child_result.fanout:
        child_result.fanout[fanout_key] = {}

    return child_result


def analyze_compute(
    node_idx, current_shape, info: AnalysisInfo
) -> SymbolicAnalysisOutput:
    einsum = info.mapping[-1].einsum
    node = info.mapping[node_idx]

    computes = 0 if info.is_copy_operation else 1

    result_accumulator = SymbolicAnalysisOutput()

    compute_key = Compute(einsum, node.component)
    result_accumulator.temporal_steps[einsum] = computes
    result_accumulator.compute_stats[compute_key] = ComputeStats(
        computes,
        computes,
        1,
    )

    if info.is_copy_operation:
        return result_accumulator

    for tensor in info.all_tensors:
        buffet = Buffet(tensor, einsum, node.component)
        stats = BuffetStats()
        stats.total_reads_to_parent = 1
        stats.max_per_parent_reads_to_parent = 1
        if tensor in info.workload.einsums[einsum].output_tensor_names:
            stats.total_writes_to_parent = 1
            stats.max_per_parent_writes_to_parent = 1
            stats.total_skipped_first_reads_to_parent = 1
            stats.min_per_parent_skipped_first_reads_to_parent = 1
        stats.max_occupancy = 1
        result_accumulator.buffet_stats[buffet] = stats

        network_node = info.job.spec.arch.find_first_of_type_above(
            NetworkSpec, node.component, default=None
        )
        if network_node is not None:
            network = Network(
                tensor,
                info.job.einsum_name,
                info.data_movement_connections.get_src(buffet),
                buffet,
                component=network_node.name if network_node else network_node,
            )
            result_accumulator.network_stats[network] = NetworkStats()

    return result_accumulator


@dataclass
class RepeatedValue[T]:
    value: T
    repeats: int


@dataclass
class SequenceOfRepatedvalues[T]:
    sequence: list[RepeatedValue[T]]


@dataclass
class StrideAndShape:
    stride: any
    shape: any


def get_stride_and_tile_shape(node: Loop, full_shape, n: int, info: AnalysisInfo):
    rank = node.rank_variable
    rank_shape = full_shape[rank]

    stride = node.tile_shape
    initial_tile_shape = node.initial_tile_shape

    # PERFECT:
    # - Node shape = stride
    # - # Iterations = total shape / stride
    # IMPERFECT:
    # - Node shape = stride
    # - # Iterations = ceil(total shape / stride)
    if IMPERFECT and initial_tile_shape is None:
        factor = sympy.ceiling(rank_shape / stride)
        stride_avg = stride / sympy.ceiling(rank_shape / stride)
        return StrideAndShape(stride_avg, RepeatedValue(stride, factor))

    if initial_tile_shape is None:
        if node._assume_perfect_factor or known_perfect_factor(stride, rank_shape):
            factor = rank_shape / stride
            return StrideAndShape(stride, RepeatedValue(stride, factor))
        else:
            factor = sympy.ceiling(rank_shape / sympy.MinGeqZero(stride, rank_shape))
            return make_possibly_different_last(stride, factor, rank_shape)

    middle_shape_factor = sympy.ceiling((rank_shape - initial_tile_shape) / stride)
    # TODO: sometimes last_shape is 0, causing numerical instability
    # Currently, we are sometimes rounding up last shape.
    # last_shape = rank_shape - initial_tile_shape - stride*middle_shape_factor
    # has_last_shape = sympy.ceiling(last_shape/(last_shape+1))
    return StrideAndShape(
        stride,
        SequenceOfRepatedvalues(
            [
                RepeatedValue(initial_tile_shape, 1),
                RepeatedValue(stride, middle_shape_factor),
                # RepeatedValue(last_shape+0.01, has_last_shape)
            ]
        ),
    )
    # if node.tile_shape is not None:
    #     tile_shape = node.tile_shape

    #     if node._assume_perfect_factor or known_perfect_factor(tile_shape, rank_shape):
    #         factor = rank_shape / tile_shape
    #         return StrideAndShape(tile_shape, RepeatedValue(tile_shape, factor))
    #     else:
    #         factor = sympy.ceiling(rank_shape / sympy.Min(tile_shape, rank_shape))
    #         return make_possibly_different_last(tile_shape, factor, rank_shape)
    # elif node.loop_bound is not None:
    #     factor = node.loop_bound

    #     if node._assume_perfect_factor or known_perfect_factor(factor, rank_shape):
    #         tile_shape = rank_shape / factor
    #         return StrideAndShape(tile_shape, RepeatedValue(tile_shape, factor))
    #     else:
    #         tile_shape = sympy.ceiling(rank_shape / sympy.Min(rank_shape, factor))
    #         return make_possibly_different_last(tile_shape, factor, rank_shape)

    # elif node.tile_pattern is not None:
    #     stride = node.tile_pattern.tile_shape
    #     initial_tile_shape = node.tile_pattern.initial_tile_shape
    #     tile_shape = node.tile_pattern.tile_shape

    #     if initial_tile_shape is not None:
    #         middle_shape_factor = sympy.ceiling((rank_shape - initial_tile_shape)/stride)
    #         # TODO: sometimes last_shape is 0, causing numerical instability
    #         # Currently, we are sometimes rounding up last shape.
    #         # last_shape = rank_shape - initial_tile_shape - stride*middle_shape_factor
    #         # has_last_shape = sympy.ceiling(last_shape/(last_shape+1))
    #         return StrideAndShape(
    #             stride,
    #             SequenceOfRepatedvalues([
    #                 RepeatedValue(initial_tile_shape, 1),
    #                 RepeatedValue(stride, middle_shape_factor),
    #                 # RepeatedValue(last_shape+0.01, has_last_shape)
    #             ])
    #         )


def known_perfect_factor(divisor, full_shape):
    return (
        isinstance(divisor, int)
        and isinstance(full_shape, int)
        and full_shape % divisor == 1
    )


def make_possibly_different_last(common_tile_shape, factor, full_shape):
    last_shape = full_shape - common_tile_shape * (factor - 1)
    all_shapes = SequenceOfRepatedvalues(
        [RepeatedValue(common_tile_shape, factor - 1), RepeatedValue(last_shape, 1)]
    )
    return StrideAndShape(common_tile_shape, all_shapes)


def insert_sympy_symbols(mapping: list[MappingNode], job: Job):
    loop_idx = 0
    symbols = []
    rank_var_with_initial = set()
    for i, node in enumerate(mapping):
        if not isinstance(node, Loop):
            continue

        stride_halos = set()
        for t in job.spec.workload.einsums[job.einsum_name].tensor_names:
            for (rank, rank_variable), (stride, halo) in job.stride_and_halo[t].items():
                if rank_variable == node.rank_variable:
                    stride_halos.add((stride, halo))

        if len(stride_halos) == 0:
            raise RuntimeError(
                f"{repr(node.rank_variable)} not found in {job.stride_and_halo}"
            )

        # We only explore imperfect for the outermost fused loops
        simple = (
            (len(stride_halos) <= 1 and next(iter(stride_halos)) == (1, 0))
            or node.rank_variable in rank_var_with_initial
            or not node._fused
        )

        # NOTE: initial_tile_shape must be inserted into `symbols` before `stride`
        # because of the order of tile shape exploration.
        # TODO: there has to be a better way to do this.
        if simple:  # Just use the stride!
            node.initial_tile_shape = None
        elif node.initial_tile_shape == SYMBOL:
            rank_var_with_initial.add(node.rank_variable)
            initial_tile_shape = sympy.symbols(
                f"initial{loop_idx}", positive=True, integer=True
            )
            symbols.append(initial_tile_shape)
            node.initial_tile_shape = initial_tile_shape

        # TODO: Check for 0 < shape < 1 for loop bound target
        if job.rank_variable_bounds[node.rank_variable] == 1:
            node.tile_shape = 1
        elif node.tile_shape == SYMBOL:
            stride = sympy.symbols(f"stride{loop_idx}", positive=True, integer=True)
            symbols.append(stride)
            node.tile_shape = stride

        # TODO: sometimes, a mapping is passed into the model twice.
        #       E.g., after calling mapper, the model is called again for more
        #       details.
        #
        # assert (
        #     node.calculated_n_iterations is None
        # ), "Number of iterations is derived from the model. Do not set it!"
        node.calculated_n_iterations = sympy.symbols(
            iterations2col(loop_idx), positive=True, integer=True
        )

        loop_idx += 1

    return symbols



"""
Flow of analysis:
-   From mapping, create the iteration space. The iteration space is the
    space of iterators in the mapping.
-   Create the relation from iteration space to operation space.
-   Create the relation from the iteration space to tensor space for each
    (buffer, tensor, einsum) tuple.
-   Run tile shape inference.

Adapted from:
https://github.com/NVlabs/timeloop/blob/4cf6d4cd043bc2a5d2eb02afa9063d7117a4dc11/ \
    src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp
Relevant Name Changes:
-   DataspaceId -> TensorName
-   LogicalBuffer -> Buffet
-   LogicalComputeUnit -> ComputeEinsum
-   Loop.op_dim -> Loop.rank_variable
-   *MappingNode.child -> MappingNode.flatten()[0]
-   Root -> Mapping
-   Compute.kernel -> Compute.einsum
-   Branch -> Split
-   FusedMapping -> Mapping
-   [node]_id -> node (conditional on MappingNode having a valid hashing functor)
"""

from collections import defaultdict, deque
from pprint import pformat
from typing import List, Optional

import islpy as isl

from accelforge.frontend.mapping import (
    Compute,
    Mapping,
    MappingNode,
    MappingNodeWithChildren,
    Sequential,
    Storage,
)
from accelforge.frontend.workload import Workload
from accelforge.frontend._workload_isl._isl import get_projection_map
from accelforge.frontend.workload import TensorName

from accelforge.model._looptree.mapping_utilities import get_paths
from accelforge.model._looptree.types import Buffet
from accelforge.model._looptree.reuse.isl.isl_functions import project_dim_in_after
from accelforge.model._looptree.reuse.isl.mapping_to_isl.skews_from_mapping import (
    skews_from_mapping,
)

from . import DUMP_ISL_IR
from .tiling import tiling_from_mapping
from .types import (
    BranchTiling,
    BufferTensorEinsum,
    ComputeEinsum,
    MappingAnalysisResult,
    Occupancy,
    OperationOccupancy,
    SkewsInfo,
)


def buffet_direct_above_sequential(mapping: Mapping) -> defaultdict[Buffet, bool]:
    """
    TODO: Verify this docstring
    For all Buffets (logical objects containing a tensor, its operating einsum,
    and a abstract hardware level), denote whether the buffet is directly above
    a :class:`~.Sequential`, or has an uninterrupted path of other buffets to a
    `~.Sequential`.

    Parameters
    ----------
    mapping:
        The mapping context of the buffets to sequential elements.

    Returns
    -------
    A dictionary of buffets and whether they're directly above a Sequential.
    """
    result: defaultdict[Buffet, bool] = defaultdict(lambda: False)
    # TODO: Figure out if get_paths is just for certain MappingNodesWithChildren
    # or not.
    for path in get_paths(mapping):
        leaf: Compute = path[-1]
        last_bufs: List[Buffet] = []

        node: MappingNode
        for node in path:
            match node:
                # If we have a storage, create a buffet for the current leaf.
                case Storage():
                    # TODO: Verify this port:
                    # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp#L518-L520
                    # Note: Buffet seems to have changed a lot?
                    # https://github.com/NVlabs/timeloop/blob/master/include/loop-analysis/isl-ir.hpp#L96
                    last_bufs.extend(
                        Buffet(tensor=tensor, einsum=leaf.einsum, level=node.component)
                        for tensor in node.tensors
                    )
                # TODO: Check that all buffets are unique, because right now
                # it seems it's dependent on the last leaf in traversal?

                # If we encounter a sequential, we know all the last buffet and its
                # parents that are buffets are directly above sequential.
                case Sequential():
                    for buf in last_bufs:
                        result[buf] = result[buf] or True
                    last_bufs.clear()
                # If we encounter no storages or a sequential, we must not be
                # directly above a sequential element, and thus can purge the path.
                case _:
                    for buf in last_bufs:
                        result[buf] = result[buf] or False
                    last_bufs.clear()

    return result


def get_parallelism(mapping: Mapping) -> defaultdict[MappingNode, float]:
    """
    Given a `accelforge.frontend.mapping.Mapping`, get the parallelism values for
    the Compute leafs.

    Parameters
    ----------
    mapping:
        The mapping to get parallelism for.

    Returns
    -------
    A map relating Compute nodes with their parallelism.
    """
    result: defaultdict[MappingNode, float] = defaultdict()

    # Initiates DFS at the root of the mapping.
    dfs_stack: deque[MappingNode] = deque([mapping])

    while dfs_stack:
        node: MappingNode = dfs_stack.pop()

        match node:
            # Recursively traverse children to find computes for parallelism.
            case MappingNodeWithChildren():
                dfs_stack.extend(node.nodes)
            # If Compute has pre-specified parallelism from internal models, trust
            # that it is right. Otherwise, assume none.
            case Compute():
                if hasattr(node, "parallelism"):
                    result[node] = node.parallelism
                else:
                    result[node] = 1
            case _:
                continue

    return result


def align_dim_names(
    map_: isl.Map,
    reference: isl.Map,
    map_align_dim_type: isl.dim_type = isl.dim_type.in_,
    reference_dim_type: Optional[isl.dim_type] = None,
) -> isl.Map:
    """
    Given an `isl.Map` and a reference `isl.Map`, align as many of the names as
    possible in the first map with the reference map.

    e.g. `map_ = [i] -> [o]` with `reference = [x] -> [y]` becomes `[x] -> [o]`
        with map_

    Parameters
    ----------
    map_:
        The map whose input is being aligned.
    reference:
        The map whose input names are used as reference for aligning `map`.
    map_align_dim_type:
        Dimension tuple in `map_` to align. Defaults to `isl.dim_type.in_`.
    reference_dim_type:
        Dimension tuple in `reference` whose names should be copied. Defaults to
        `map_align_dim_type`.

    Returns
    -------
    A version of `map_` with aligned input names.
    """
    if reference_dim_type is None:
        reference_dim_type = map_align_dim_type

    for dim_idx in range(
        min(map_.dim(map_align_dim_type), reference.dim(reference_dim_type))
    ):
        dim_name: Optional[str] = reference.get_dim_name(reference_dim_type, dim_idx)
        if dim_name is not None:
            map_ = map_.set_dim_name(map_align_dim_type, dim_idx, dim_name)

    return map_


def occupancies_from_mapping(
    mapping: Mapping, workload: Workload
) -> MappingAnalysisResult:
    """
    Given a Mapping and a Workload, extract the data occupancies in memory.

    Parameters
    ----------
    mapping:
        The Mapping of data to hardware.
    workload:
        The Workload occurring on chip.

    Returns
    -------
    The occupancies as an analysis of the Workload on Mapping.
    """
    branch_tiling: BranchTiling = tiling_from_mapping(mapping, workload)
    # tiling: [tile_iteration_space] -> [iteration_space]
    if DUMP_ISL_IR:
        for node, tiling in branch_tiling.items():
            print(f"[Tiling]Node({node}): {tiling}")
            # TODO: Port this line
            # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp#L55-L64
            print(f"[Ops]Node({node}): ")

    occupancies: defaultdict[BufferTensorEinsum, Occupancy] = defaultdict()
    skews: SkewsInfo = skews_from_mapping(mapping, workload)
    # skew [Spacetime] -> [tile_iteration_space]
    if DUMP_ISL_IR:
        print(f"skews: {pformat(skews)}")

    ### Somewhere, call the domain space of the returned tilings {einsum}_iterations
    for bte, skew in skews.bte_to_skew.items():
        if DUMP_ISL_IR:
            print(f"{bte} has skew: {skew}")
        tiling = branch_tiling[bte.einsum]

        accesses: Optional[isl.Map] = None
        read_tensors: set[TensorName] = workload.einsums[
            bte.einsum.einsum
        ].input_tensor_names
        write_tensors: set[TensorName] = workload.einsums[
            bte.einsum.einsum
        ].output_tensor_names

        if bte.tensor in read_tensors or bte.tensor in write_tensors:
            accesses = get_projection_map(
                workload.einsums[bte.einsum.einsum], bte.tensor
            )
        else:
            continue

        aligned_skew: isl.Map = align_dim_names(skew.map_, tiling)
        if DUMP_ISL_IR:
            print(f"Skew: {skew.map_}")
            print(f"Aligned Skew: {aligned_skew}")
            print(f"Tiling: {tiling}")
            print(f"{tiling.apply_range(accesses)}")
            print(f"{skew.map_.dim(isl.dim_type.out)}")
        occupancy: isl.Map = aligned_skew.apply_range(
            project_dim_in_after(
                tiling.apply_range(accesses),
                skew.map_.dim(isl.dim_type.out),
                # TODO: fix this unsafe mixing.
            ).set_tuple_name(
                isl.dim_type.in_, aligned_skew.get_tuple_name(isl.dim_type.out)
            )
        )

        occupancies[bte] = Occupancy(skew.tags, occupancy)

    operations_occupancies: defaultdict[ComputeEinsum, OperationOccupancy] = (
        defaultdict()
    )
    for ce, skew in skews.ce_unit_to_skew.items():
        tiling: isl.Map = branch_tiling[ce.branch_leaf_node]
        if DUMP_ISL_IR:
            print(f"skew.map_ {skew.map_}")
        operation_occupancy: isl.Map = skew.map_.apply_range(
            project_dim_in_after(
                tiling, skew.map_.dim(isl.dim_type.out)
            ).set_tuple_name(
                # TODO: Unify the names at some point...
                isl.dim_type.in_,
                skew.map_.get_tuple_name(isl.dim_type.out),
            )
        )
        operations_occupancies[ce] = OperationOccupancy(skew.tags, operation_occupancy)

    return MappingAnalysisResult(
        buffet_to_occupancy=occupancies,
        compute_einsum_to_occupancy=operations_occupancies,
        buffet_direct_above_sequential=buffet_direct_above_sequential(mapping),
        compute_to_assumed_parallelism=get_parallelism(mapping),
        branch_tiling=branch_tiling,
    )



"""
File for all the functions that conduct tiling analysis for the overall mapping
analysis.
"""

from collections import defaultdict, deque
from typing import List, Tuple, Optional

from pprint import pformat

import islpy as isl

from accelforge.frontend.mapping import (
    # Types
    MappingNode,
    # Mapping objects
    Mapping,
    MappingNodeWithChildren,
    Nested,
    # Physical object types in Mappings.
    Compute,
    Storage,
    # Logical object types in Mappings.
    Loop,
    Spatial,
    Temporal,
    Split,
)
from accelforge.frontend.workload import (
    # Workload class for all of AccelForge.
    Workload,
)
from accelforge.frontend._workload_isl._isl import (
    get_einsum_operation_space,
    get_projection_map,
)
from accelforge.frontend.mapping import TensorName
from accelforge.model._looptree.reuse.isl.isl_functions import (
    add_dims_preserve_name_map,
    insert_dims_preserve_name_map,
    map_to_prior_coordinate,
)
from accelforge.model._looptree.reuse.isl.mapping_to_isl import DUMP_ISL_IR
from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import (
    EinsumName,
    Tiling,
    BranchTiling,
)


def get_mapping_group_einsums(
    mapping: Mapping,
) -> defaultdict[MappingNode, set[EinsumName]]:
    """
    From a mapping, get the group of einsums for a given node.

    Parameters
    ----------
    mapping:
        The mapping we are getting the grouped einsums for.

    Returns
    -------
    A dictionary relating a MappingNode to a set of einsums.
    """
    # Each pair is a (current_node, last_non_branch_node)
    dfs_stack: deque[Tuple[MappingNode, MappingNode]] = deque()
    # Each pair is a (last_non_branch_node, set_of_children_nodes)
    child_stack: deque[Tuple[MappingNode, set[MappingNode]]] = deque()
    result: defaultdict[MappingNode, set[EinsumName]] = defaultdict(set)

    # Start DFS hierarchical search from the root.
    dfs_stack.append((mapping, mapping))

    # Exhaustive DFS search.
    while dfs_stack:
        # Grabs latest node to search.
        node, last_non_branch = dfs_stack.pop()

        # Differentiates behavior by number of child nodes.
        match node:
            case MappingNodeWithChildren():
                match len(node.nodes):
                    # No children, log as a folded result.
                    case 0:
                        # Note:: Check necesary in case Distrobuffers elides
                        # computes into one large unit.
                        if isinstance(node, Compute):
                            result[last_non_branch].add(node.einsum)
                        else:
                            raise TypeError(
                                f"The following node should be of class "
                                f"Compute as it has no children:\n---\n{node}"
                            )
                    # Explore the children further.
                    case 1:
                        dfs_stack.append((node.nodes[0], last_non_branch))
                    # Log all branching children and explore all children.
                    case _:
                        children: set[MappingNode] = set(node.nodes)
                        child_stack.append((last_non_branch, children))
                        dfs_stack.extend((child, child) for child in children)
            # Assumed no children, log as a folded result.
            case Compute():
                result[last_non_branch].add(node.einsum)
            # These had children in Timeloop we had to add to the DFS, but because
            # of our extension of dfs_stack we can just skip this node.
            case Spatial() | Temporal() | Storage():
                continue
            case _:
                raise AttributeError(
                    f"The following node of class {type(node)} has "
                    f"indeterminant number of children:\n---\n"
                    f"{node}"
                )

    # Push up einsums to parents.
    for node, children in reversed(child_stack):
        node_einsum_set: set[EinsumName] = result[node]
        for child in children:
            node_einsum_set.update(result[child])

    return result


def get_head_among_einsums(
    einsum_set: set[EinsumName], workload: Workload
) -> set[EinsumName]:
    """
    Gets the provider einsums that only consume data (i.e., sink einsums).

    Parameters
    ----------
    einsum_set:
        Set of einsums to consider.
    workload:
        The workload context the einsums exist in.

    Returns
    -------
    The set of all head einsums.
    """
    # Returns set of einsums that are not data producers.
    return {
        einsum
        for einsum in einsum_set
        if all(
            not any(
                consumer.name in einsum_set
                for consumer in workload.einsums_with_tensor_as_input(output_tensor)
            )
            for output_tensor in workload.einsums[einsum].output_tensor_names
        )
    }


def add_new_tile_dim(
    old_tiling: Tiling, dim_idx: int, tile_size: int, rank_var: Optional[str] = None
) -> Tiling:
    """
    Given a tiling, add a new dimension to the tiling.

    Parameters
    ----------
    old_tiling:
        The previous tiling the mapper proposed.
    dim_idx:
        The index of the dimension being tiled.
    tile_size:
        The size of the tiling on dim_idx.
    rank_var:
        Rank variable name to assign to the new input dimension, if provided.

    Returns
    -------
    The new Tiling with tiled dimension at dim_idx.
    """

    # new_tiling has one extra dimension at the end compared to old_tiling.
    new_tiling = insert_dims_preserve_name_map(
        old_tiling, isl.dim_type.in_, old_tiling.dim(isl.dim_type.in_), 1
    )
    if rank_var:
        new_tiling = new_tiling.set_dim_name(
            isl.dim_type.in_, old_tiling.dim(isl.dim_type.in_), rank_var
        )

    # Min and max of dim_idx. dimension being tiled as function of tiled dimensions.
    dim_min: isl.PwAff = new_tiling.dim_min(dim_idx)
    dim_max: isl.PwAff = new_tiling.dim_max(dim_idx)

    # Aff from tiled dimensions space to value of newest dim.
    new_dim_id: isl.Aff = isl.Aff.var_on_domain(
        dim_min.get_domain_space().to_local_space(),
        isl.dim_type.set,
        dim_min.dim(isl.dim_type.in_) - 1,
    )

    # Aff from tiled dimensions space to tile tile size constant.
    tile_size_aff: isl.Aff = isl.Aff.val_on_domain_space(
        dim_min.get_domain_space(), isl.Val.int_from_ui(isl.DEFAULT_CONTEXT, tile_size)
    )

    # PwAff from tiled dimension space to tile_size * newest_dim.
    tile_translate: isl.PwAff = isl.PwAff.from_aff(new_dim_id.mul(tile_size_aff))

    # What dim_min should be given new tiling.
    new_dim_min: isl.PwAff = dim_min.add(tile_translate)

    # What dim_max should be given new tiling.
    new_dim_max: isl.PwAff = new_dim_min.add(
        isl.PwAff.from_aff(tile_size_aff.add_constant_val(-1))
    )

    # TODO: Might be logically equivalent to new_dim_id:
    # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/tiling.cpp#L52-L59
    new_iter_id: isl.PwAff = isl.PwAff.from_aff(
        isl.Aff.var_on_domain(
            new_tiling.get_space().domain(),
            isl.dim_type.set,
            old_tiling.dim(isl.dim_type.in_),
        )
    )

    # The set of valid values of the new tiled dimensions.
    iter_set: isl.Set = new_tiling.domain()
    iter_set = iter_set.intersect(new_iter_id.le_set(dim_max.div(tile_size_aff).ceil()))
    iter_set = iter_set.intersect(new_dim_min.ge_set(dim_min))

    # The value of iter dims cannot exceed what was available before tiling.
    new_tiling = new_tiling.intersect_domain(iter_set)

    # The set of operations need to to follow the new tile bounds.
    identity: isl.PwAff = isl.PwAff.from_aff(
        isl.Aff.var_on_domain(new_tiling.get_space().range(), isl.dim_type.set, dim_idx)
    )
    new_tiling = new_tiling.intersect(new_dim_min.le_map(identity))
    new_tiling = new_tiling.intersect(new_dim_max.ge_map(identity))

    return new_tiling


def shared_input_based_tile_shape_inference(
    workload: Workload,
    tiling_info: defaultdict[EinsumName, Tiling],
    einsums: set[EinsumName],
    shared_input_tensor: TensorName,
    tiled_einsum: EinsumName,
) -> None:
    """
    Given a `tiled_einsum` in a `workload`, restrict the other `einsums`' execution
    in this tiling to one in which the data is shared with the `tiled_einsum`. This
    is because, when tiled, data is multicast so the other einsums being tiled together
    must shared data.

    Parameters
    ----------
    workload:
        The workload context the tiling is occurring in.
    tiling_info:
        Relation of `EinsumName` and its viable tiling on hardware.
    einsums:
        The set of all einsums.
    shared_input_tensor:
        The singular tensor `einsums` all read from.
    tiled_einsum:
        The einsum being tiled.

    Returns
    -------
    None

    Postconditions
    --------------
    `tiling_info` is updated such that each Tiling contains only compatible tilings
    with `tiled_einsum`.
    """
    # Gets the data tiled_einsum reads from shared_input_tensor
    tiled_einsum_read_accesses: isl.Map = get_projection_map(
        workload.einsums[tiled_einsum], shared_input_tensor
    )
    read_data: isl.Map = tiling_info[tiled_einsum].apply_range(
        tiled_einsum_read_accesses
    )

    # Goes through all other einsums and restrict their tilings to only the executable
    # operations after one of the einsums is tiled.
    for einsum in einsums:
        if einsum == tiled_einsum:
            continue

        read_accesses: isl.Map = get_projection_map(
            workload.einsums[einsum], shared_input_tensor
        )
        executable_operations: isl.Map = read_data.apply_range(read_accesses.reverse())
        executable_operations = executable_operations.intersect_range(
            get_einsum_operation_space(workload, einsum)
        )

        tiling_info[einsum] = tiling_info[einsum].intersect(executable_operations)


def consumer_based_tile_shape_inference(
    workload: Workload,
    tiling_info: defaultdict[EinsumName, Tiling],
    tensor_to_reuse_level: defaultdict[TensorName, int],
    einsums: set[EinsumName],
    tiled_einsum: EinsumName,
):
    """
    Given a `tiled_einsum` in a `workload`, restrict the other `einsums`' execution
    in this tiling to one in which the data is required for the tensors read by
    `tiled_einsum`. This is because, when tiled, data is multicast so the other
    einsums being tiled together must shared data.

    Parameters
    ----------
    workload:
        The workload context the tiling is occurring in.
    tiling_info:
        Relation of `EinsumName` and its viable tiling on hardware.
    tensor_to_reuse_level:
        A relation between a tensor and the amount of reuse occurring.
    einsums:
        The set of all einsums.
    tiled_einsum:
        The einsum being tiled.

    Returns
    -------
    None

    Postconditions
    --------------
    `tiling_info` is updated such that each Tiling contains only compatible tilings
    with `tiled_einsum`.
    """
    # Goes recursively through tensor dependencies (read tensors) and tiles them.
    queue: deque[EinsumName] = deque([tiled_einsum])
    while queue:
        einsum: EinsumName = queue.popleft()
        tiling: Tiling = tiling_info[einsum]

        # For each tensor read by this einsum, tile that tensor's producers.
        for tensor in workload.einsums[einsum].input_tensor_names:
            producer_einsums: set[EinsumName] = {
                e for e in workload.einsums[einsum].output_tensor_names
            }
            if len(producer_einsums) > 1:
                raise NotImplementedError(
                    "Tile shape inference cannot handle multiple einsums writing the same tensor."
                )

            # Not an intermediate tensor.
            if not producer_einsums:
                continue

            producer_einsums.intersection_update(einsums)
            # No producer einsum in this fusion set.
            if not producer_einsums:
                continue

            # Collates all the consumer einsum read accesses.
            producer_einsum: EinsumName = next(iter(producer_einsums))
            read_accesses: isl.Map = get_projection_map(
                workload.einsums[einsum], tensor
            )
            # Required data of the tiling as a mapping of read accesses.
            required_data: isl.Map = tiling.apply_range(read_accesses)

            # Calculates the data computed by the producer einsums.
            computed_data: isl.Map = required_data
            if tensor in tensor_to_reuse_level:
                reuse_level: int = tensor_to_reuse_level[tensor]
                shifter: isl.Map = map_to_prior_coordinate(
                    tiling.dim(isl.dim_type.in_),
                    reuse_level,
                    tiling.get_tuple_name(isl.dim_type.in_),
                )
                buffered_data: isl.Map = shifter.apply_range(required_data)
                computed_data = computed_data.subtract(buffered_data).coalesce()

            # Grabs the elements this tensor relies on from producer_einsums.
            producer_write_dependency: isl.Map = get_projection_map(
                workload.einsums[producer_einsum], tensor
            )
            # Gets the required operations to produce the current tensor.
            required_operations: isl.Map = computed_data.apply_range(
                producer_write_dependency.reverse()
            )
            required_operations = required_operations.intersect_range(
                get_einsum_operation_space(workload, producer_einsum)
            )

            # Mutations of the tilings of producer einsums.
            # TODO: Deal with fusing naming better (perhaps mix the names?)
            tiling_info[producer_einsum] = tiling_info[producer_einsum].intersect(
                required_operations.set_tuple_name(
                    isl.dim_type.in_,
                    tiling_info[producer_einsum].get_tuple_name(isl.dim_type.in_),
                )
            )

            queue.append(producer_einsum)


def detect_shared_input_tensor(
    fused_set: set[EinsumName], workload: Workload
) -> List[TensorName]:
    """
    Given a set of fused einsums on a workload, detect the input tensor that they
    all are dependent on, if it exists.

    Parameters
    ----------
    fused_set:
        The set of fused einsums being analyzed.
    workload:
        The workload context the einsums exist in.

    Returns
    -------
    The list of tensors shared by the inputs. Because we default to consumer-based
    analysis if there's more than 1 shared input among the tensors, we only return
    tuple sizes of {0, 1, 2}.
    """
    n_einsums: int = 0
    tensor_read_counts: defaultdict[TensorName, int] = defaultdict(lambda: 0)

    # Counts the number of times a tensor is read by an einsum.
    for einsum in fused_set:
        for tensor in workload.einsums[einsum].input_tensor_names:
            tensor_read_counts[tensor] += 1
        n_einsums += 1

    shared_input_tensors: List[TensorName] = []
    for tensor, count in tensor_read_counts.items():
        # Tensor is shared by all einsums.
        if count == n_einsums:
            shared_input_tensors.append(tensor)
            # Caller should resort to consumer-based fusing methods.
            if len(shared_input_tensors) > 1:
                return shared_input_tensors

    return shared_input_tensors


def tiling_from_mapping(mapping: Mapping, workload: Workload) -> BranchTiling:
    """
    Given a mapping and a workload generates a tiling.

    Parameters
    ----------
    mapping:
        A mapping of data to hardware.
    workload:
        The problem being solved.

    Returns
    -------
    BranchTiling associating a node's ID with its tiling.
    """
    result: BranchTiling = BranchTiling()
    # Grabs the head einsums.
    mapping_groups: defaultdict[MappingNode, set[EinsumName]] = (
        get_mapping_group_einsums(mapping)
    )
    mapping_group_heads: defaultdict[MappingNode, set[EinsumName]] = defaultdict(
        set,
        {
            node: get_head_among_einsums(group, workload)
            for node, group in mapping_groups.items()
        },
    )

    tensor_to_reuse_level: defaultdict[TensorName, int] = defaultdict()
    dfs_stack: deque[MappingNode] = deque([mapping])  # DFS starts at mapping root.

    # Maps last non-branch to tiling of each in the group.
    tiling_info: defaultdict[MappingNode, defaultdict[EinsumName, Tiling]] = (
        defaultdict(defaultdict)
    )

    # Appends info for the root.
    for einsum_name in workload.einsum_names:
        tiling_info[mapping][einsum_name] = isl.Map.from_range(
            get_einsum_operation_space(workload, einsum_name)
        ).set_tuple_name(isl.dim_type.in_, f"{einsum_name}_tiled_iteration")

    # Tracks rank_var specified to partitioned_rank_var index, as traversal
    # in tiling goes down the partition.
    rank_var_partitions: defaultdict[str, int] = defaultdict(lambda: 0)

    def _get_rank_var_partition(rank_var: str) -> str:
        """
        Given a rank_var, get the partition at the current point in execution
        and increment for the next retrieval.
        """
        nonlocal rank_var_partitions
        rank_var_partition: str = f"{rank_var}{rank_var_partitions[rank_var]}"
        rank_var_partitions[rank_var] += 1
        return rank_var_partition

    def _tile_branch(heads: set[EinsumName], fusing_node: MappingNode):
        """
        Given a set of `heads` to fuse at `fusing_node`, fuse as much as possible
        in this branch.

        Parameters
        ----------
        heads:
            The heads being fused.
        fusing_node:
            The node node in the mapping at which the fusing is happening.

        Preconditions
        -------------
        1. `dfs_stack`: initialized with tiles to proceed to explore.
        2. `tiling_info`: prima facie populated.
        3. `tensor_to_reuse_level`: initialized and unmutated from last time this
            function was run.

        Postconditions
        --------------
        1. `dfs_stack`: progressed to the next node to tile at.
        2. `tiling_info`: updated to include the fusing and tiling.
        3. `tensor_to_reuse_level`: populated if information has changed from tiling.
        """
        nonlocal dfs_stack
        nonlocal tiling_info
        nonlocal tensor_to_reuse_level

        current_node: MappingNode = fusing_node
        while True:
            # Fuses current_node to one of the heads.
            match current_node:
                # For or Par-For loop handling.
                case Loop():
                    if len(heads) != 1:
                        raise ValueError(
                            f"Cannot fuse tiled set with {len(heads)} heads.\n"
                        )

                    # Tiles `current_node.rank_variable` at `head`
                    head = next(iter(heads))
                    tiling: Tiling = tiling_info[fusing_node][head]
                    # Downstreams of "heads" is also constant as it is a set, not
                    # AbstractSet.
                    idx: int = tuple(workload.einsums[head].rank_variables).index(
                        current_node.rank_variable
                    )

                    # Adds a new tile_dim to the old tiling.
                    # TODO: Handle stride.
                    if (
                        isinstance(
                            _ := current_node.tile_pattern.initial_tile_shape, int
                        )
                        and (_ != 0)
                        and (_ == current_node.tile_pattern.tile_shape)
                    ):
                        tiling: Tiling = add_new_tile_dim(
                            tiling,
                            idx,
                            current_node.tile_pattern.initial_tile_shape,
                            _get_rank_var_partition(current_node.rank_variable),
                        )
                    else:
                        raise NotImplementedError(
                            f"Tile size analysis not implemented for type {type(fusing_node)} "
                            f"with tile shape {current_node.tile_pattern.initial_tile_shape}"
                        )

                    # Saves the fused tiling.
                    tiling_info[fusing_node][head] = tiling

                    # Adds the ranks to the tiling isl.Map.
                    iteration_set: isl.Set = tiling.domain()
                    for einsum in mapping_groups[fusing_node] - {head}:
                        tiling = tiling_info[fusing_node][einsum]
                        # Index variables for the branch.
                        tiling = insert_dims_preserve_name_map(
                            tiling, isl.dim_type.in_, tiling.dim(isl.dim_type.in_), 1
                        )
                        tiling = tiling.set_dim_name(
                            isl.dim_type.in_,
                            tiling.dim(isl.dim_type.in_) - 1,
                            _get_rank_var_partition(current_node.rank_variable),
                        )
                        # TODO: Figure out if this intersection is correct.
                        tiling = tiling.intersect_domain(
                            iteration_set.set_tuple_name(
                                tiling.get_tuple_name(isl.dim_type.in_)
                            )
                        )
                        tiling_info[fusing_node][einsum] = tiling

                    current_node = dfs_stack.pop()
                # Notes what reuse level the tensor is on.
                case Storage():
                    # See current_node is the highest level of Storage to determine reuse level.
                    for tensor in current_node.tensors:
                        # Check second term
                        if tensor not in tensor_to_reuse_level:
                            random_einsum: EinsumName = next(
                                iter(mapping_groups[fusing_node])
                            )
                            tiling: Tiling = tiling_info[fusing_node][random_einsum]
                            tensor_to_reuse_level[tensor] = tiling.dim(isl.dim_type.in_)

                    current_node = dfs_stack.pop()
                # If we are at the Mapping root, just go to the actual Nodes.
                case Mapping():
                    dfs_stack.extend(reversed(current_node.nodes))
                    current_node = dfs_stack.pop()
                # If we hit the compute node, we've finished tiling, end!
                case Compute():
                    result[current_node] = tiling_info[fusing_node][current_node.einsum]
                    return
                case Split():
                    fused_set: set[EinsumName] = mapping_groups[fusing_node]
                    if len(heads) != 1:
                        # There can't be a tiling, so no inference to be done.
                        break

                    random_head = next(iter(heads))
                    if len(_ := detect_shared_input_tensor(fused_set, workload)) == 1:
                        shared_input_based_tile_shape_inference(
                            workload,
                            tiling_info[fusing_node],
                            fused_set,
                            _[0],
                            random_head,
                        )
                    else:
                        consumer_based_tile_shape_inference(
                            workload,
                            tiling_info[fusing_node],
                            tensor_to_reuse_level,
                            fused_set,
                            random_head,
                        )

                    # Goes through each child node of the current node and propagate
                    # the tiling updates.
                    for idx, child in enumerate(current_node.nodes):
                        # Each child needs tilings for all Einsums in its group.
                        group: set[EinsumName] = mapping_groups[child]
                        tilings: defaultdict[EinsumName, Tiling] = defaultdict()

                        # For all einsums the child is involved in, update their tilings.
                        for einsum in group:
                            tiling: Tiling = tiling_info[fusing_node][einsum]
                            # Add dimension that iterates over branches.
                            new_tiling: Tiling = add_dims_preserve_name_map(
                                tiling, isl.dim_type.in_, 1
                            )

                            tilings[einsum] = new_tiling.fix_input_si(
                                new_tiling.dim(isl.dim_type.in_) - 1, idx
                            )

                        # Update the tiling info for the child.
                        tiling_info[child] = tilings
                        # DFS tile on the child.
                        dfs_stack.append(child)

                    return
                case Nested():
                    dfs_stack.extend(reversed(current_node.nodes))
                    current_node = dfs_stack.pop()
                case _:
                    raise NotImplementedError(
                        f"Type {type(fusing_node)} not handled.\n"
                        f"---\n"
                        f"node={pformat(fusing_node)}"
                    )

    while dfs_stack:
        fusing_node = dfs_stack.pop()
        if DUMP_ISL_IR:
            print(f"New Tiling Root: {pformat(fusing_node)}")
        _tile_branch(mapping_group_heads[fusing_node], fusing_node)

    return result
